"""Collection of utility functions."""

import cobra
import logging

import memote.support.helpers as helpers
from memote.utils import truncate

from libsbml import Reaction
from six import iteritems
from typing import Union

__author__ = "Gwendolyn O. Döbel and Carolin Brune"

################################################################################
# variables
################################################################################

# SBO terms
# ---------
SBO_BIOCHEM_TERMS = ["SBO:0000377", "SBO:0000399", "SBO:0000402", "SBO:0000403",
                   "SBO:0000660", "SBO:0000178", "SBO:0000200", "SBO:0000214",
                   "SBO:0000215", "SBO:0000217", "SBO:0000218", "SBO:0000219",
                   "SBO:0000220", "SBO:0000222", "SBO:0000223", "SBO:0000233",
                   "SBO:0000376", "SBO:0000401"] #: :meta: 

# compartments
# ------------
VALID_COMPARTMENTS = {'c': 'cytosol', 'e': 'extracellular space', 'p':'periplasm','y':'unknown compartment'} #: :meta: 
COMP_MAPPING = {'c': 'c', 'e': 'e', 'p': 'p',
                'C_c': 'c', 'C_e': 'e', 'C_p': 'p',
                '':'y'} #: :meta:


################################################################################
# functions
################################################################################

# @DISCUSSION: Might be useful for logging?
##### Coloured text:
# from colorama import init as colorama_init
# from colorama import Fore
# 
# def coloured_example_text():
#     colorama_init(autoreset=True)
#     print(f'{Fore.RED}To use the KEGG comparison the specification of the organismid (KEGG organism code) is obligatory.\n' +
#       'If there is no organism code available for your organism in KEGG but an entry for your organism exists in BioCyc, use the option \'BioCyc\'.\n' +
#       'If no entry for your organism exists in KEGG and/or BioCyc, the gap analysis cannot be done.')

# SBO
# ---

# @TODO extend to cover more reactions
def reannotate_sbo_memote(model:cobra.Model) -> cobra.Model:
   """Reannotate the SBO annotations (e.g. from SBOannotator) of a model 
   into the SBO scheme accessible by memote.

   Args:
      - model (cobra.Model): 
         The cobra Model to be reannotated.

   Returns:
       cobra.Model: 
         The reannotated model 
   """

   # biochem reactions
   for r in model.reactions:
      if 'sbo' in r.annotation:
         if r.annotation['sbo'] in SBO_BIOCHEM_TERMS:
            r.annotation['sbo'] = "SBO:0000176"

    # @TODO: add transport reactions?


# handling biomass reaction
# -------------------------
def test_biomass_presence(model: cobra.Model) -> Union[list[str], None]:
    """
    Modified from MEMOTE: https://github.com/opencobra/memote/blob/81a55a163262a0e06bfcb036d98e8e551edc3873/src/memote/suite/tests/test_biomass.py#LL42C3-L42C3
    
    Expect the model to contain at least one biomass reaction.

    The biomass composition aka biomass formulation aka biomass reaction
    is a common pseudo-reaction accounting for biomass synthesis in
    constraints-based modelling. It describes the stoichiometry of
    intracellular compounds that are required for cell growth. While this
    reaction may not be relevant to modeling the metabolism of higher
    organisms, it is essential for single-cell modeling.

    Implementation:
    Identifies possible biomass reactions using two principal steps:
    
        1. Return reactions that include the SBO annotation "SBO:0000629" for
        biomass.
        
        2. If no reactions can be identified this way:
        
            1. Look for the ``buzzwords`` "biomass", "growth" and "bof" in reaction IDs.
            2. Look for metabolite IDs or names that contain the ``buzzword`` "biomass" and obtain the set of reactions they are involved in.
            3. Remove boundary reactions from this set.
            4. Return the union of reactions that match the buzzwords and of the reactions that metabolites are involved in that match the buzzword.
        
    This test checks if at least one biomass reaction is present.
    
    If no reaction can be identified return None.

    """
    biomass_rxn = [rxn.id for rxn in helpers.find_biomass_reaction(model)]
    outcome = len(biomass_rxn) > 0
    logging.info(
        """In this model the following {} biomass reaction(s) were
        identified: {}""".format(
            len(biomass_rxn), truncate(biomass_rxn)
        )
    )
    if outcome: return biomass_rxn
    else: return None


def sum_biomass_weight(reaction: Reaction) -> float:
    """
    From MEMOTE: https://github.com/opencobra/memote/blob/81a55a163262a0e06bfcb036d98e8e551edc3873/src/memote/support/biomass.py#L95
    
    Compute the sum of all reaction compounds.

    This function expects all metabolites of the biomass reaction to have
    formula information assigned.

    Args:
        - reaction (Reaction): 
            The biomass reaction of the model under investigation.

    Returns:
        float: 
            The molecular weight of the biomass reaction in units of g/mmol.
    """
    return (
        sum(
            -coef * met.formula_weight
            for (met, coef) in iteritems(reaction.metabolites)
        )
        / 1000.0
    )
    
    
def test_biomass_consistency(model: cobra.Model, reaction_id: str) -> Union[float, str]:
    """
    Modified from MEMOTE: https://github.com/opencobra/memote/blob/81a55a163262a0e06bfcb036d98e8e551edc3873/src/memote/suite/tests/test_biomass.py#L89
    
    Expect biomass components to sum up to 1 g[CDW].

    This test only yields sensible results if all biomass precursor
    metabolites have chemical formulas assigned to them.
    The molecular weight of the biomass reaction in metabolic models is
    defined to be equal to 1 g/mmol. Conforming to this is essential in order
    to be able to reliably calculate growth yields, to cross-compare models,
    and to obtain valid predictions when simulating microbial consortia. A
    deviation from 1 - 1E-03 to 1 + 1E-06 is accepted.

    Implementation:
    Multiplies the coefficient of each metabolite of the biomass reaction with
    its molecular weight calculated from the formula, then divides the overall
    sum of all the products by 1000.

    Args: 
        - model(cobraModel):
            The model loaded with COBRApy.
        - reaction_id(str):
            Reaction ID of a BOF.

    Returns:
        (1) Case: problematic input
                str: 
                    an error message.
        
        (2) Case: successful testing
                float: 
                    biomass weight
    """
    reaction = model.reactions.get_by_id(reaction_id)
    try:
        biomass_weight = sum_biomass_weight(reaction)
    except TypeError:
        message = """
        One or more of the biomass components do not have a defined formula or contain unspecified chemical groups.
        The biomass overall weight could thus not be calculated.
        """
        return message
    else:
        if ((1 - 1e-03) < biomass_weight < (1 + 1e-06)):
            logging.info(            
                """The component molar mass of the biomass reaction {} sums up to {}
                which is inside the 1e-03 margin from 1 mmol / g[CDW] / h.
                """.format(
                    reaction_id, biomass_weight
                    )
                )
        else:
            logging.warning(
                """The component molar mass of the biomass reaction {} sums up to {}
                which is outside of the 1e-03 margin from 1 mmol / g[CDW] / h.
                """.format(
                    reaction_id, biomass_weight
                )
            )
    #outcome = (1 - 1e-03) < biomass_weight < (1 + 1e-06) -> Need to implement that for check
    # To account for numerical inaccuracies, a range from 1-1e0-3 to 1+1e-06
    # is implemented in the assertion check
    return biomass_weight



#!/usr/bin/env python
"""Provides the medium class and additional functions to handle media.

Functionalities include (amongst others):

- loading a medium into a Medium object from a database, a file or a model
- adding a medium to a model
- adding media information to the database
- extending, change and manipulate various parts of a medium to create the desired medium 
"""

__author__ = "Carolin Brune"

################################################################################
# requirements
################################################################################

import cobra
import copy
import io
import numpy as np
import pandas as pd
import random
import re
import sqlite3
import string
import sys
import warnings

from colorama import init as colorama_init
from colorama import Fore
from pathlib import Path
from sqlite_dump import iterdump
from typing import Literal, Union, Any

from ..utility.databases import PATH_TO_DB, PATH_TO_DB_FOLDER

############################################################################
# variables
############################################################################

ALLOWED_DATABASE_LINKS = ['BiGG', 'MetaNetX', 'SEED', 'VMH', 'ChEBI', 'KEGG'] #: :meta: 
REQUIRED_SUBSTANCE_ATTRIBUTES = ['name', 'formula', 'flux', 'source'] #: :meta: 
INTEGER_REGEX = re.compile('^[-+]?([1-9]\d*|0)$') #: :meta: 
FLOAT_REGEX = re.compile(r'[+-]?(\d+(\.\d*)?|\.\d+)([eE][+-]?\d+)?') #: :meta:  

############################################################################
# classes
############################################################################


class Medium:
    """Class describing a medium.

    Attributes:
        - name (str): 
            The name or abbreviation of the medium.
        - substance_table (pd.DataFrame): 
            A table containing information about the medium in silico compounds. Long format.
        - description (str, optional): 
            Short description of the medium. Defaults to None.
        - doi (str): 
            Reference(s) to the original publication of the medium. Defaults to None.
    """

    def __init__(self, name:str, substance_table:pd.DataFrame=pd.DataFrame(columns=['name','formula','flux','source','db_id','db_type']), description:str=None, doi:str=None):
        """Initialise a Medium object.

        Args:
            - name (str): 
                The name or abbreviation of the medium.
            - substance_table (pd.DataFrame, optional):  
                A table containing information about the medium in silico compounds. Long format.
                Defaults to an empty table with the columns ['name','formula','flux','source','db_id','db_type'].
            - description (str, optional): 
                Short description of the medium. Defaults to None.
            - doi (str, optional): 
                Reference(s) to the original publication of the medium.. Defaults to None.
        """
        
        self.name = name
        self.description = description
        self.substance_table = substance_table
        self.doi = doi

    def add_substance_from_db(self, name:str, flux:float=10.0):
        """Add a substance from the database to the medium.

        Args:
            - name (str): 
                Name of the substance. Should be part of the database substance.name column.
            - flux (float, optional): 
                Sets the flux value of the new substance. Defaults to 10.0.
        """
        # build connection to DB
        connection = sqlite3.connect(PATH_TO_DB)
        cursor = connection.cursor()
        # fetch substance
        result = cursor.execute(""" SELECT substance.name, substance.formula, substance2db.db_id, substance2db.db_type 
                                    FROM substance, substance2db 
                                    WHERE substance.name = ? AND substance.id = substance2db.substance_id""", (name,))
        substance = result.fetchall()

        # check if fetch was successful
        if len(substance) == 0:
            warnings.warn(f'Could not fetch substance {name} from DB.')
            return
        
        # add substance to table
        substance_table = pd.DataFrame(substance, columns=['name','formula','db_id','db_type'])
        substance_table.insert(2,'flux',flux)
        substance_table.insert(3,'source',None)
        self.substance_table = pd.concat([self.substance_table, substance_table], ignore_index=True)

    # @TEST
    # @IDEA : more options, e.g. allow different conditions for removal (flux, formula, db_id or multi)
    def remove_substance(self, name: str):
        """Remove a substance from the medium based on its name

        Args:
            - name (str): 
                Name of the substance to remove.
        """
        
        self.substance_table = self.substance_table[self.substance_table['name'] != name]
    

    def get_source(self, element:str) -> list[str]:
        """Get the source of a given element for the medium.

        Search for the given element (elemental symbol e.g. O), excluding pattern matches that are 
        followed by other lower-case letters and returm them as a list of sources for the given element. 

        Args:
            - element (str): 
                The symbol of the element to search the sources of

        Returns:
            list[str]: 
                The list of the names of the sources (no duplicates).
        """

        return list(set(self.substance_table[self.substance_table['formula'].str.contains(element + '(?![a-z])', case=True, regex=True).fillna(False)]['name']))


    # @TEST
    # @ASK : is deleting ALL other sources to rigid?
    def set_source(self, element:str, new_source:str):
        """Set the source for a given element to a specific substance by deleting all
        other sources of said element before adding the new source.

        Args:
            - element (str): 
                The element to set the source for, e.g. 'O' for oxygen.
            - new_source (str): 
                The new source. Should be the name of a substance in the database, otherwise no new source will be set.
        """

        # get sources for element
        current_sources = self.get_source(element)
        # remove current sources from medium
        for s in current_sources:
            self.remove_substance(s)
        # add new source to medium
        self.add_substance_from_db(new_source)


    def is_aerobic(self) -> bool:
        """Check if the medium contains O2 / dioxygen.

        Returns:
            bool: 
                Results of the test, True if pure oxygen is in the medium.
        """

        test = self.substance_table[['name','formula']].loc[(self.substance_table['name'] == 'Dioxygen [O2]') & (self.substance_table['formula'] == 'O2')].any().all()
        return test
    
    
    def make_aerobic(self, flux:float=None):
        """If the medium is curretly anaerobic, add oxygen to the medium to make it aerobic.
        
        Args:
            - flux(float,optional): 
                The flux value for the oxygen to be added. Defaults to None.
        """
        if not self.is_aerobic():
            # build connection to DB
            connection = sqlite3.connect(PATH_TO_DB)
            cursor = connection.cursor()
            # fetch oxygen
            result = cursor.execute(""" SELECT substance.name, substance.formula, substance2db.db_id, substance2db.db_type 
                                    FROM substance, substance2db 
                                    WHERE substance.formula = 'O2' AND substance.name = 'Dioxygen [O2]' AND substance.id = substance2db.substance_id""")
            oxygen = result.fetchall()
            # add oxygen to substance table
            oxygen_table = pd.DataFrame(oxygen, columns=['name','formula','db_id','db_type'])
            oxygen_table.insert(2,'flux',flux)
            oxygen_table.insert(3,'source',None)
            self.substance_table = pd.concat([self.substance_table, oxygen_table], ignore_index=True)


    def make_anaerobic(self):
        """If the medium is currently aerobic, deletes the oxygen from it to make it anaerobic.
        """
        if self.is_aerobic():
            # remove dioxygen // O2 from the substance table
            self.substance_table.drop(self.substance_table[(self.substance_table['name']=='Dioxygen [O2]') & (self.substance_table['formula']=='O2')].index, inplace=True)

    # @IDEA: provide dict with names and fluxes to set some of them separatly if user wishes to do so - or separate function?
    def set_default_flux(self,flux:float =10.0, replace:bool =False, double_o2:bool =True):
        """Set a default flux for the model.

        Args:
            - flux (float, optional): 
                Default flux for the medium. Defaults to 10.0.
            - replace (bool, optional): 
                Replace al fluxes with the default.
                Defaults to False.
            - double_o2 (bool, optional): 
                Tag to double the flux for oxygen only. Works only with replace=True. Defaults to True.
        """
        # replace fluxes
        if replace:
            self.substance_table['flux'] = flux
            if self.is_aerobic() and double_o2:
                self.substance_table.loc[self.substance_table['name']=='Dioxygen [O2]','flux'] = 2*flux
        # keep already set fluxes
        else:
            self.substance_table['flux'] = self.substance_table['flux'].fillna(flux)
            if self.is_aerobic() and double_o2 and self.substance_table.loc[self.substance_table['name']=='Dioxygen [O2]','flux'].any():
                self.substance_table.loc[self.substance_table['name']=='Dioxygen [O2]','flux'] = 2*flux
        
    def set_oxygen_percentage(self, perc:float=1.0):
        """Set oxygen percentage of the medium.

        Args:
            - perc (float, optional): 
                Percentage of oxygen. Defaults to 1.0 (= 100%)
        """

        if self.is_aerobic():
            self.substance_table.loc[self.substance_table['name']=='Dioxygen [O2]','flux'] = self.substance_table.loc[self.substance_table['name']=='Dioxygen [O2]','flux'] * perc
        else:
            warnings.warn(f'WARNING: no oxygen detected in medium {self.name}, cannot set oxygen percentage.')


    def combine(self,other:'Medium', how:Union[Literal['+'],None,float,tuple[float]]='+', default_flux:float=10.0, inplace:bool=False) -> 'Medium':
        """Combine two media into a new one.

        Modes to combine media (input for param how):

        - None -> combine media, remove all flux values (= set them to None). Sets sources to None as well.
        - '+' -> Add fluxes of the same substance together.
        - float -> Calculate flux * percentage (float) for first medium and flux * 1.0-percentage (float) for second medium and add fluxes of same substance together.
        - tuple(float,float) -> Same as above, except both percentages are given.

        Args:
            - other (Medium): 
                The medium to combine with.
            - how (Union[Literal['+'],None,float,tuple[float]], optional): 
                How to combine the two media. 
                Options listed in header. Defaults to '+'.
            - default_flux (float, optional): 
                Flux to use in combine-modes (except how=None) for NaN/None values. 
                Defaults to 10.0.

        Returns:
            Medium: 
                The combined medium.
        """

        def combine_media_with_fluxes(combined:'Medium', second_medium:'Medium') -> 'Medium':
            """Helper function for :py:func:`combine`. Add a substance table to a medium by
            combining fluxes of substances of the same name.

            Args:
                - combined (Medium): 
                    The medium that will be the combined one.
                - second_medium (Medium): 
                    The substance table to integrate into the medium.

            Returns:
                Medium: 
                    The combined medium.
            """
            # get dict of name + flux of second medium
            possible_dups = second_medium[['name','flux']].groupby('name').first().reset_index()
            possible_dups = dict(zip(possible_dups.name, possible_dups.flux))
            # search for overlapping substances and add fluxes together
            added = []
            for subs, flux in possible_dups.items():
                if subs in combined.substance_table['name'].values:
                    combined.substance_table.loc[combined.substance_table['name'] == subs,'flux'] += flux
                    added.append(subs)
            # add remaining substances to medium from the second one
            second_medium = second_medium[~second_medium['name'].isin(added)]
            combined.substance_table = pd.concat([combined.substance_table, second_medium], ignore_index=True)

            return combined

        # combine description and Co
        if inplace:
            combined = self 
        else:
            combined = copy.deepcopy(self)
        combined.name = self.name + '+' + other.name
        combined.description = f'Combined medium constructed from {self.name} and {other.name}.'
        combined.doi = str(self.doi) + ', ' + str(other.doi)

        # combine substance table
        match how:

            # add fluxes together
            case '+':
                # replace None values
                combined.substance_table['flux'] = combined.substance_table['flux'].fillna(default_flux)
                to_add = copy.deepcopy(other.substance_table)
                to_add['flux'] = to_add['flux'].fillna(default_flux)
                # add together
                combined = combine_media_with_fluxes(combined, to_add)

            # add fluxes based on percentages
            case float() | tuple():
                if type(how) == float:
                    how = (how, 1.0-how)
                # replace None values
                combined.substance_table['flux'] = combined.substance_table['flux'].fillna(default_flux)
                second_medium = copy.deepcopy(other.substance_table)
                second_medium['flux'] = second_medium['flux'].fillna(default_flux)
                # calculate new fluxes based on percentages
                combined.substance_table['flux'] = combined.substance_table.flux.apply(lambda x: x*how[0])
                second_medium['flux'] = second_medium.flux.apply(lambda x: x*how[1])
                # add together
                combined = combine_media_with_fluxes(combined, second_medium)

            # combine and set all fluxes to 0
            case None:
                combined.substance_table = pd.concat([combined.substance_table, other.substance_table], ignore_index=True)
                # remove fluxes and sources, as they are no longer a fit
                combined.substance_table['flux'] = None
                combined.substance_table['source'] = None
                # remove duplicate rows
                combined.substance_table.drop_duplicates(inplace=True,ignore_index=True)

            case _:
                raise ValueError(f'Unknown input for parameter how: {how}')

        return combined
    
    
    def __add__(self,other:'Medium') -> 'Medium':
        return self.combine(other)

    def add_subset(self, subset_name:str, default_flux:float=10.0, inplace:bool=True) -> 'Medium':
        """Add a subset of substances to the medium, returning a newly generated one.

        Args:
            - subset_name (str): 
                The type of subset to be added. Name should be in database-substset-id.
            - default_flux (float, optional): 
                Default flux value to calculate fluxes from based  on the percentages saved in the database. 
                Defaults to 10.0. 

        Returns:
            Medium: 
                A new medium that is the combination of the set subset and the old one.
                In the case that the given subset name is not found in the database, 
                the original medium is returned.
        """

        # open connection to database
        connection = sqlite3.connect(PATH_TO_DB)
        cursor = connection.cursor()

        # check if subset name is valid
        result = cursor.execute("""SELECT 1 FROM subset WHERE subset.name = ?""",(subset_name,))
        check = result.fetchone()
        if check:

            # retrieve subset from database
            db_res = cursor.execute("""SELECT substance.name, substance.formula, subset2substance.percent, substance2db.db_id, substance2db.db_type
                                    FROM substance, subset, subset2substance, substance2db
                                    WHERE subset.name = ? AND subset.id = subset2substance.subset_id AND subset2substance.substance_id = substance.id AND substance.id = substance2db.substance_id
                                    """,(subset_name,))
            subs = db_res.fetchall()

            # reformat
            # --------
            # rename
            subs = pd.DataFrame(subs,columns = ['name', 'formula', 'percent', 'db_id', 'db_type'])
            # percent -> flux
            subs['flux'] = subs['percent'].apply(lambda x: default_flux*x)
            subs.drop('percent', axis=1, inplace=True)
            # add source column
            subs['source'] = None
            # sort
            colorder = ['name', 'formula', 'flux', 'source', 'db_id', 'db_type']
            subs = subs.reindex(columns=colorder)
    
            # create subset
            sub_medium = Medium(subset_name, subs, description=f'subset {subset_name}')

            # combine with current
            return self.combine(sub_medium, inplace=inplace)

        else:
            warnings.warn(f'Could not find subset in DB, nothing added to medium: {subset_name}')
            # just return the original medium
            return self


    # functions for export table
    # --------------------------

    # @TODO
    def format_substance_table(self, format:str) -> pd.DataFrame:
        """Produce a reformatted version of the substance table for different purposes.

        Possible formats

        - 'growth_old': for working with the old version of the growth module.

        Args:
            - format (str): 
                Specifies the output format type.

        Raises:
            - ValueError: Unknown format type for table.

        Returns:
            pd.DataFrame: 
                The reformatted substance table (copy).
        """

        match format:
            # for enabling usage of this new medium class with the old growth functions
            case 'growth_old':

                formatted_table = self.substance_table.loc[self.substance_table['db_type']=='BiGG'][['name','flux','db_id']]
                formatted_table.rename(columns={'db_id':'BiGG'})

                # ..............................................................
                # TODO / WARNING
                # currently only substances WITH a BiGG ID are kept in this step
                # ..............................................................
                
                formatted_table['mediumname'] = self.name

            # raise error for unknow input
            case _:
                raise ValueError(f'Unknown format type for table: {format}')
            
        return formatted_table
    

    def produce_medium_docs_table(self, folder: str = './', max_width: int = 80) -> str:
        """Produces a rst-file containing reStructuredText for the substance table for documentation.

        Args:
            - folder (str, optional): 
                Path to folder/directory to save the rst-file to. Defaults to './'.
            - max_width (int, optional): 
                Maximal table width of the rst-table. Defaults to 80.
        """

        def calculate_column_widths_docs(header: list, max_width: int, flux_width = 8) -> str:
            """Helper function for :py:func:`produce_medium_docs_table`. 
            Calculates the columns widths based on the header lengths and maximal widths.

            Args:
                - header (list): 
                    List of column names of the table.
                - max_width (int): 
                    Maximal widths of the output table.
                - flux_width (int, optional): 
                    Widths of the flux column, if 'flux' is part of 'header'. Defaults to 8.
 
            Returns:
                str: The columns widths as a string, separated by a whitespace.
            """

            if len(header) == 2:
                partition = max_width // 3
                return f"{str(max_width-partition)} {partition}"
            else:
                partition = (max_width - flux_width) // 2
                return f"{str(max_width-flux_width-partition)} {flux_width} {partition}"

        def produce_medium_docs_table_row(row: pd.Series, file: io.TextIOWrapper):
            """Helper function for :py:func:`produce_medium_docs_table`.
            Tranforms each row of the substance table into a row of the rst-file.

            Args:
                - row (pd.Series): 
                    The row of the Medium.substance_table.
                - file (io.TextIOWrapper): 
                    The connection to the file to write the rows into.
            """

            list = row.to_list()
            file.write(f"  * - {list[0]}\n")
            for l in list[1:]:
                file.write(f"    - {l}\n")

        # make sure given directory path ends with '/'
        if not folder.endswith('/'):
            folder = folder + '/'

        with open(folder + f'{self.name}.rst', 'w') as f:

            # slim table to columns of interest for documentation
            m_subs = self.substance_table[['name','flux','source']]
            m_subs.drop_duplicates(keep='first', inplace=True)

            if all(m_subs['flux'].values == None):
                m_subs = m_subs.drop('flux', axis=1)

            header = m_subs.columns

            widths = calculate_column_widths_docs(header, max_width)
            
            title = f'{self.description} ({self.name})'
            
            # Produce header/title of HTML page
            f.write(
                f"{title}\n"
                f"{'^' * len(title)}\n\n"
            )

            # produce descriptor
            f.write(f".. list-table:: {self.name} composition\n"
                    f"  :name: {self.name.lower()}_comp\n"
                    "  :align: center\n"
                    f"  :widths: {widths}\n"
                    "  :header-rows: 1\n"
                    "  :class: no-scrollbar-table\n\n")
            
            # produce header
            f.write(f"  * - {header[0]}\n")
            for l in header[1:]:
                f.write(f"    - {l}\n")

            # produce table body
            m_subs.apply(produce_medium_docs_table_row, file=f, axis=1)

            f.close()


    def export_to_file(self, type:str='tsv',dir:str='./', max_widths: int = 80):
        """Export medium, especially substance table.

        Args:
            - type (str, optional): 
                Type of file to export to. Defaults to 'tsv'. Further choices are 'csv', 'docs', 'rst'.
            - dir (str, optional): 
                Path to the directory to write the file to. Defaults to './'.
            - max_widths (int, optional): 
                Maximal table width for the documentation table (). Only viable for 'rst' and 'docs'.
                Defaults to 80.

        Raises:
            - ValueError: Unknown export type if type not in ['tsv','csv','docs','rst']
        """

        match type:
            case 'tsv':
                self.substance_table.to_csv(Path(dir,self.name + '_min_medium' + '.tsv'), sep='\t', index=False)
            case 'csv':
                self.substance_table.to_csv(Path(dir,self.name + '_min_medium' + '.csv'), sep=';', index=False)
            case 'docs' | 'rst':
                self.produce_medium_docs_table(folder = dir, max_width = max_widths)
            case _:
                raise ValueError('Unknown export type: {type}')
            

    # functions for conversion
    # ------------------------
    # @TODO
    #    extend namespace options
    #        maybe a warning or something similar, if compounds do not have an ID corresponding to the namespace
    # @TEST
    #    name option for namespace
    # @DEV : extent literals in all related functions after extension of namespace options
    def export_to_cobra(self, namespace:Literal['Name', 'BiGG']='BiGG', default_flux:float=10.0, replace:bool=False, double_o2:bool=True) -> dict[str,float]:
        """Export a medium to the COBRApy format for a medium.

        Args:
            - namespace (Literal['Name', 'BiGG'], optional): 
                Namespace to use. Defaults to 'BiGG'.
            - default_flux (float, optional): 
                Default flux to substitute missing values. Defaults to 10.0.
            - replace (bool, optional): 
                Replace all values with the default flux. Defaults to False.
            - double_o2 (bool, optional): 
                Double the flux of oxygen. Defaults to True.

        Raises:
            - ValueError: Unknown namespace.

        Returns:
            dict[str,float]: 
                The exported medium.
        """
        
        self.set_default_flux(default_flux, replace=replace, double_o2=double_o2)

        match namespace:
            case 'Name':

                names = self.substance_table[['name','flux']]
                names['EX'] = biggs['db_id'] + ' exchange'
                cobra_medium = pd.Series(names.flux.values,index=names.EX).to_dict()       

            case 'BiGG':
                
                biggs = self.substance_table[self.substance_table['db_type'].str.contains('BiGG')][['name','db_id','flux']]
                biggs['db_id_EX'] = 'EX_' + biggs['db_id'] + '_e'
                cobra_medium = pd.Series(biggs.flux.values,index=biggs.db_id_EX).to_dict()       

                # .....................................
                # TODO
                #    more options
                # .....................................       

            case _:
                raise ValueError(f'Unknown namespace: {namespace}')
        
        return cobra_medium  

############################################################################
# functions for loading and writing from DB
############################################################################

# @TODO also used in Medium class, but put in there as a subfunction
    # - solution?
def produce_docs_table_row(row: pd.Series, file: io.TextIOWrapper):

        list = row.to_list()
        file.write(f"  * - {list[0]}\n")
        for l in list[1:]:
            file.write(f"    - {l}\n")


def load_substance_table_from_db(mediumname: str, database:str, 
                                 type:Literal['testing','standard']='standard') -> pd.DataFrame:
    """Load a substance table from a database.

    Currently available types:

    - 'testing': for debugging
    - 'standard': The standard format containing all information in long format.

    Note: 'documentation' currently object to change

    Args:
        - name (str): 
            The name (or identifier) of the medium.
        - database (str): 
            Path to the database.
        - type (Literal['testing','standard'], optional):
            How to load the table. Defaults to 'standard'.

    Raises:
        - ValueError: Unknown type for loading the substance table.

    Returns:
        pd.DataFrame: 
            The substance table in the specified type retrieved from the database.
    """
    
    # build connection to DB
    connection = sqlite3.connect(database)
    cursor = connection.cursor()

    match type:
        # use this for debugging
        case 'testing':
            result = cursor.execute("""SELECT substance.id, substance.name, substance.formula, medium2substance.flux 
                                    FROM medium, medium2substance, substance
                                    WHERE medium.name = ? AND medium.id = medium2substance.medium_id AND medium2substance.substance_id = substance.id
                                    """, (mediumname,)) 
            substance_table = result.fetchall()
            substance_table = pd.DataFrame(substance_table, columns=['id','name','formula','flux'])

        # create table with all information, standard for generating the Medium table
        case 'standard':
            result = cursor.execute("""SELECT substance.name, substance.formula, medium2substance.flux , medium2substance.source, substance2db.db_id, substance2db.db_type
                                    FROM medium, medium2substance, substance, substance2db
                                    WHERE medium.name = ? AND medium.id = medium2substance.medium_id AND medium2substance.substance_id = substance.id AND substance2db.substance_id = substance.id
                                    """, (mediumname,)) 
            substance_table = result.fetchall()
            substance_table = pd.DataFrame(substance_table, columns=['name','formula','flux','source','db_id','db_type'])

        # default: throw error
        case _:
            connection.close()
            raise ValueError(f"Unknown type for loading the substance table: {type}")

    # close connection
    connection.close()

    return substance_table


def load_medium_from_db(name:str, database:str=PATH_TO_DB, type:str='standard') -> Medium:
    """Load a medium from a database.

    Args:
        - name (str): 
            The name (or identifier) of the medium.
        - database (str, optional): 
            Path to the database. Defaults to the in-built database.
        - type (str, optional): 
            How to load the medium. Defaults to 'standard'.

    Raises:
        - ValueError: Unknown medium name.

    Returns:
        Medium: 
            The medium retrieved from the database.
    """
    
    # build connection to DB
    connection = sqlite3.connect(database)
    cursor = connection.cursor()

    # get description
    result = cursor.execute("SELECT medium.description FROM medium WHERE medium.name = ?",(name,))
    if result:
        description = result.fetchone()[0] # each name should be unique
    else:
        raise ValueError(f'Unknown medium name: {name}')

    # get DOI(s)
    result = cursor.execute("SELECT medium.reference FROM medium WHERE medium.name = ?",(name,))
    doi = result.fetchone()[0]

    # close connection to DB
    connection.close()
    
    # get substance table
    substance = load_substance_table_from_db(name, database, type)

    return Medium(name=name, substance_table=substance, description=description, doi=doi)


def load_subset_from_db(subset_name:str) -> tuple[str,str,pd.DataFrame]:
    """Load a subset from the database.
    
    Args:
        - subset_name(str): 
            Name of the subset to be loaded.
    
    Returns:
        tuple of (1) str, (2) str and (3) pd.DataFrame
            (1) name of the subset
            (2) description of the subset
            (3) substance table for the subset
    """
    
    # open connection to database
    connection = sqlite3.connect(PATH_TO_DB)
    cursor = connection.cursor()

    # check if subset name is valid
    result = cursor.execute("""SELECT * FROM subset WHERE subset.name = ?""",(subset_name,))
    check = result.fetchone()
    if check:
        # set name 
        name = subset_name
        # set description
        description = check[2]
        # retrieve subset from database
        db_res = cursor.execute("""SELECT substance.name, subset2substance.percent
                                FROM substance, subset, subset2substance
                                WHERE subset.name = ? AND subset.id = subset2substance.subset_id AND subset2substance.substance_id = substance.id
                                """,(subset_name,))
        substance_table = pd.DataFrame(db_res.fetchall(), columns=['name','percent'])

    return (name, description, substance_table)


def generate_docs_for_subset(subset_name:str, folder:str='./', max_width:int=80):
    """Generate documentation for a subset.

    Args:
        - subset_name (str): 
            Name of the subset.
        - folder (str, optional): 
            Folder to save the output to. Defaults to './'.
        - max_width (int, optional): 
            Maximal table width for the documentation page. Defaults to 80.
    """
    
    name, description, subs = load_subset_from_db(subset_name)

    with open(Path(folder,f'{name}.rst'), 'w') as f:
        
        partition = max_width // 3
        width = f"{str(max_width-partition)} {partition}"
        header = subs.columns

        # Produce header/title of HTML page
        f.write(
            f"{name}\n"
            f"{'^' * len(name)}\n"
        )

        f.write(description + '\n\n')

        # produce descriptor
        f.write(f".. list-table:: {name} composition\n"
                f"  :name: {name.lower()}_comp\n"
                "  :align: center\n"
                f"  :widths: {width}\n"
                "  :header-rows: 1\n"
                "  :class: no-scrollbar-table\n\n")
        
        # produce header
        f.write(f"  * - {header[0]}\n")
        for l in header[1:]:
            f.write(f"    - {l}\n")

        # produce table body
        subs.apply(produce_docs_table_row, file=f, axis=1)

    
############################################################################
# functions for reading media from extern
############################################################################

def read_substances_from_file(path: str) -> pd.DataFrame: 
    """Read in a TSV with substance information into a table.

    Format of the TSV (with example):
    name \t formula \t flux \t source \t X \t X | ...
    water \t H20 \t 10.0 \t .....

    X: placeholder for database names (columns filled with corresponding IDs of the substances)
    X = see ALLOWED_DATABASE_LINKS

    Args:
        - path(str): 
            The path to the input file.

    Returns:
        pd.DataFrame: 
            The table of substance information read from the file
    """

    # read in the table
    substance_table = pd.read_csv(path, sep='\t', comment='#')

    # validate the table
    substance_cols = substance_table.columns
    for req in REQUIRED_SUBSTANCE_ATTRIBUTES:
        if req in substance_cols:
            continue
        else:
            substance_table[req] = None

    # remove unwanted database links
    all_allowed = REQUIRED_SUBSTANCE_ATTRIBUTES + ALLOWED_DATABASE_LINKS
    for_removal = [_ for _ in substance_cols if _ not in all_allowed] 
    substance_table.drop(for_removal, axis=1, inplace=True)

    # convert all NaNs into None
    substance_table = substance_table.replace({np.nan: None})

    return substance_table

# @TEST file flag for how
def read_external_medium(how:Literal['file','console'], **kwargs) -> Medium:
    """Read in an external medium. 

    Currently available options for how

    - 'console': read in medium via the console
    - 'file': read in medium from a file, requires a 'path=str' argument to be passed.

    About the format (console, file):

    The substances have to be saved in a TSV file table (format see :py:func:`read_substances_from_file`).
    Further information for the 'file' option have to added as comments in the format: `# info_name: info`.
    Information should but do not need to contain name, description and reference.

    Args:
        - how (Literal['file','console']): 
            How (or from where) the medium should be read in.
            Available options are given above.

    Raises:
        - ValueError: Unknown description for how.

    Returns:
        Medium: 
            The read-in medium.
    """

    match how:
        # interactive mode using the console
        case 'console':

            # prompt for name
            name = input('Enter a medium name:\n')

            # prompt for description
            description = input('Enter a short description for the medium:\n')

            # promt for description 
            reference = input('Enter the reference link (DOI) for the medium:\n')

            # prompt for path to substance table
            substance_path = input('Enter a path to a table of substances (TSV):\n')
            # tsv format
            substances = read_substances_from_file(substance_path)

            # construct medium
            new_medium = Medium(name=name, description=description, substance_table=substances, doi=reference)

            return new_medium
        
        # read from a file 
        case 'file':

            # get info from comments
            infos = {}
            with open(kwargs['path'], 'r') as f:
                for line in f:
                    # parse the comment lines from the file
                    if line.startswith('#'):
                        l = line.split(':')
                        k = l[0]
                        k = k[1:].strip()
                        v = "".join(l[1:])
                        v = v.strip()
                        infos[k] = v
                    else:
                        break
            # retrieve important infos and supplement missing
            if 'name' in infos.keys():
                name = infos['name']
            else:
                warnings.warn('No name found for externally loaded medium. Setting random one.')
                name = 'noname_' + ''.join(random.choices(string.ascii_uppercase + string.digits, k=6))
            
            if 'description' in infos.keys():
                description = infos['description']
            else:
                description = None
            
            if 'reference' in infos.keys():
                reference = infos['reference']
            else:
                reference = None

            # get substances
            substances = read_substances_from_file(kwargs['path'])

            # construct medium
            new_medium = Medium(name=name, description=description, substance_table=substances, doi=reference)

            return new_medium

        # unknown case, raise error
        case _:
            raise ValueError(f'Unknown input for parameter how: {how}')


def extract_medium_info_from_model_bigg(row, model:cobra.Model) -> pd.Series:
    """Helper function for :py:func:`read_from_cobra_model`. 
    Extracts more information about the medium.

    Args:
        - row (pd.Series): 
            A row of the datatable of :py:func:`read_from_cobra_model`.
        - model (cobra.Model): 
            The cobra Model

    Returns:
        pd.Series: 
            One row for the substance table.
    """

    db_id = row['db_id_EX'].replace('EX_','').replace('_e','')
    meta = model.metabolites.get_by_id(db_id+'_e')
    row['name'] = meta.name
    row['formula'] = meta.formula
    row['db_id'] = db_id

    return row


def read_from_cobra_model(model: cobra.Model, namespace:Literal['BiGG']='BiGG') -> Medium:
    """Read and import a medium from a cobra model into a Medium object.

    Args:
        - model (cobra.Model): 
            An open cobra Model.

    Returns:
        Medium: 
            The imported medium.
    """

    # retrieve the medium from the model
    cobra_medium = model.medium.copy()
    substances = pd.DataFrame(cobra_medium.items(),columns=['db_id_EX','flux'])
    substances['db_type'] = namespace

    # retrieve additional information
    substances['source'] = model.id
    substances['name'] = None
    substances['formula'] = None

    # transform exchange reacs into the metabolite 
    match namespace:
        case 'BiGG':    
            substances = substances.apply(extract_medium_info_from_model_bigg, model=model, axis=1)
            #substances['db_id_meta'] = substances.db_id_EX.apply(lambda x: x.removeprefix('EX_'))
            #substances['db_id'] = substances.db_id_meta.apply(lambda x: x.rsplit('_',maxsplit=1)[0])
        case _:
            raise ValueError(f'Unknown input for namespace: {namespace}')
        

    # reformat table 
    substances.drop(columns=['db_id_EX'],inplace=True)
    sorted_cols = ['name','formula','flux','source','db_id','db_type']
    substances = substances.reindex(columns=sorted_cols)

    # construct the medium 
    name = f'medium_{model.id}'
    description = f'Medium imported from model {model.id}'
    imported_medium = Medium(name, substances, description)

    return imported_medium

############################################################################
# functions for adding entries to DB
############################################################################

# utility
# -------

def get_last_idx_table(tablename: str, connection: sqlite3.Connection, cursor: sqlite3.Cursor) -> int:
    """Helper function.
    Retrieves the last row id of a specified table of the database.

    Args:
        - tablename (str): 
            The name of the table to retrieve the last row id from
        - connection (sqlite3.Connection): 
            Connection to the database.
        - cursor (sqlite3.Cursor): 
            Cursor for the database.

    Returns:
        int: 
            The last row ID of the specified table.
    """

    match tablename:
        case 'medium':
            last_rowid_res = cursor.execute("SELECT MAX(rowid) FROM medium")
        case 'medium2substance':
            last_rowid_res = cursor.execute("SELECT MAX(rowid) FROM medium2substance")
        case 'substance':
            last_rowid_res = cursor.execute("SELECT MAX(rowid) FROM substance")
        case 'substance2db':
            last_rowid_res = cursor.execute("SELECT MAX(rowid) FROM substance2db")
        case _:
            sys.exit(f'Unknown table name for database: {tablename}')

    last_rowid_fetched = last_rowid_res.fetchone()
    last_rowid = last_rowid_fetched[0]

    return last_rowid


# add a new medium
# ----------------

def enter_substance_row(row: pd.Series, connection: sqlite3.Connection, cursor:sqlite3.Cursor) -> int:
    """Helper function for :py:func:`~refinegems.classes.medium.enter_medium_into_db`.
    Enters a new entry in the medium2substance table.

    Args:
        - row (pd.Series): 
            A row of the pd.DataFrame of the :py:func:`~refinegems.classes.medium.enter_medium_into_db` function.
        - connection (sqlite3.Connection): 
            Connection to the database.
        - cursor (sqlite3.Cursor): 
            Cursor for the database.

    Returns:
        int: 
            The substance ID in the database of the substance.
    """

    # check if a perfect match exists in database
    exact_match_res = cursor.execute("SELECT * FROM substance WHERE substance.formula = ? AND substance.name = ? ",(row['formula'],row['name']))
    exact_match = exact_match_res.fetchone()

    if exact_match:

        # get substance id
        substance_id = exact_match[0]

    # if not promt for similar or create new substance
    else:    

        candidates = cursor.execute("SELECT * FROM substance WHERE substance.formula = ? OR substance.name LIKE ? ",(row['formula'],row['name']))
        candidate_list = candidates.fetchall()

        # if candidates were found, let user choose if one of those should be used
        
        substance_id = None
        if len(candidate_list) > 0:

            print(f'Following similar entries have been found for the substance {row["name"]}.')
            for i in candidate_list:
                print(i)
            res = input('If one matches your entry, please enter the ID or enter skip/s: \n')
            while True:
                # skip
                if res in ['skip','s']:
                    break
                # match found
                elif res.isnumeric() and int(res) in [_[0] for _ in candidate_list]:
                    substance_id = int(res)
                    break
                # unknown input, try again
                else:
                    res = input('Unknown input. Please try again:\n')

        # if no matching entry has been found, insert new entry into DB
        if not substance_id:
            cursor.execute("INSERT INTO substance VALUES(?,?,?)",(None,row['name'],row['formula'],))
            connection.commit()
            # retrieve the newly added ID
            res = cursor.execute("SELECT last_insert_rowid() FROM substance")
            substance_id = res.fetchone()[0]
        
    return substance_id


def enter_m2s_row(row: pd.Series, medium_id: int, connection: sqlite3.Connection, cursor: sqlite3.Cursor):
    """Helper function for :py:func:`refinegems.classes.medium.enter_medium_into_db`.
    Enters a new entry in the medium2substance table.

    Args:
        - row (pd.Series): 
            A row of the pd.DataFrame of the :py:func:`~refinegems.classes.medium.enter_medium_into_db` function.
        - medium_id (int): 
            The row id of the medium.
        - connection (sqlite3.Connection): 
            Connection to the database.
        - cursor (sqlite3.Cursor): 
            Cursor for the database.
    """

    # check if entry already exists in database
    exact_match_res = cursor.execute("SELECT 1 FROM medium2substance WHERE medium2substance.medium_id = ? AND medium2substance.substance_id = ? ",(medium_id,row['substance_id']))
    exact_match = exact_match_res.fetchone()

    # else
    if not exact_match:
        cursor.execute('INSERT INTO medium2substance VALUES(?,?,?,?)',(medium_id,row['substance_id'],row['flux'],row['source'],))
        connection.commit()
    else:
        warnings.warn(f'Medium2substance connection {medium_id} - {row["substance_id"]} already exists, skipped second assignment.')
    

def enter_s2db_row(row: pd.Series, db_type: str, connection: sqlite3.Connection, cursor: sqlite3.Cursor):
    """Helper function for :py:func:`~refinegems.classes.medium.enter_medium_into_db`.
    Enters a new entry in the substance2db table after checking if it has yet to be added.

    Args:
        - row (pd.Series): 
            A row of the pd.DataFrame of the :py:func:`~refinegems.classes.medium.enter_medium_into_db` function.
        - db_type (str): 
            Type of database identifier to be added.
        - connection (sqlite3.Connection): 
            Connection to the database.
        - cursor (sqlite3.Cursor): 
            Cursor for the database.
    """

    # only need to enter if ID exists
    if row[db_type]:
        # check if mapping already exists
        entry_check_res = cursor.execute("SELECT 1 FROM substance2db WHERE substance2db.substance_id = ? AND db_id = ?", (row['substance_id'],row[db_type],))
        entry_check = entry_check_res.fetchone()
        if not entry_check:
            cursor.execute('INSERT INTO substance2db VALUES(?,?,?)',(row['substance_id'],row[db_type],db_type,))
            connection.commit()


def enter_medium_into_db(medium: Medium, database: str= PATH_TO_DB):
    """Enter a new medium to an already existing database.

    Args:
        - medium (Medium): 
            A medium object to be added to the database.
        - database (str, optional): 
            Path to the database. Defaults to the in-build databse.
    """

    # build connection to DB
    connection = sqlite3.connect(database)
    cursor = connection.cursor()

    # collect starting IDs
    sid_medium = get_last_idx_table('medium', connection, cursor)
    sid_medium2substance = get_last_idx_table('medium2substance', connection, cursor)
    sid_substance = get_last_idx_table('substance', connection, cursor)
    sid_substance2db = get_last_idx_table('substance2db', connection, cursor)

    # try to add a new medium to the database
    try:

        # name
        #   check, if name already in DB and prompt, if that problem occurs
        #   enter into database, if no problem remains 
        name_check_res = cursor.execute("SELECT 1 FROM medium WHERE medium.name = ?", (medium.name,))
        name_check = name_check_res.fetchone()
        if not name_check:
            # medium name is unique 
            # add medium to DB
            cursor.execute("INSERT INTO medium VALUES(?,?,?,?)",(None,medium.name,medium.description,medium.doi,))
        else:
            # medium name already in DB 
            check = input('The name {medium.name} is already in the database.\nDo you want to set a new name? (yes/no): ')
            if check in ['y','yes']:
                # set a new name
                while True:
                    new_name = input('Please enter a medium name:\n')
                    name_check_2 = cursor.execute("SELECT 1 FROM medium WHERE medium.name = ?", (new_name,))
                    if name_check_2.fetchone():
                        print('This name already exists in the database.')
                    else:
                        medium.name = new_name
                        break
            elif check in ['n','no']:
                # end program when no new name is set
                print('No new name chosen. Ending the program.')
                sys.exit()
            else:
                # Abort program at unknown input
                sys.exit('Unknown input. Aborting.')
        
        connection.commit()
        res = cursor.execute("SELECT last_insert_rowid() FROM medium")
        medium_id = res.fetchone()[0]

        # substances 
        medium.substance_table['substance_id'] = medium.substance_table.apply(enter_substance_row, axis=1, args=(connection, cursor))

        # medium2substances
        medium.substance_table.apply(enter_m2s_row, axis=1, args=(medium_id, connection, cursor))

        # substance2db
        avail_data = [_ for _ in medium.substance_table.columns if _ in ALLOWED_DATABASE_LINKS]
        for current_db in avail_data:
            medium.substance_table.apply(enter_s2db_row, axis=1, args=(current_db,connection,cursor))

        print(f'Medium {medium.name} with ID {medium_id} has been added to the database.')

    # in case of problems, interupt and revert changes
    except:

        print(f'During execution, the following error occured: \n{sys.exc_info()}')
        print('Reverting changes to database...')

        cursor.execute("DELETE FROM medium WHERE rowid > ?",(sid_medium,))
        cursor.execute("DELETE FROM medium WHERE rowid > ?",(sid_medium2substance,))
        cursor.execute("DELETE FROM medium WHERE rowid > ?",(sid_substance,))
        cursor.execute("DELETE FROM medium WHERE rowid > ?",(sid_substance2db,))

        print('Done')

    # in any case close connection to database
    finally:
        connection.close()


# add a new subset
# ----------------
        
def add_subset_to_db(name:str, desc:str, subs_dict:dict, 
                     database: str= PATH_TO_DB, default_perc:float=1.0) -> None:
    """Add a new subset to the database.

    Args:
        - name (str): 
            Name (Abbreviation) of the new subset. Needs to be unique for the databse.
        - desc (str): 
            Description of the new subset.
        - subs_dict (dict): 
            Dictionary of the names and percentages for the substances to be included
            in the new subsets. The names should be part of the substance table.
        - database (str, optional): 
            Which database to connect to. 
            Defaults to PATH_TO_DB.
        - default_perc (float, optional): 
            Default percentage to set if None is given in the dictionary. 
            Defaults to 1.0.
    """

    
    # build connection to DB
    connection = sqlite3.connect(database)
    cursor = connection.cursor()

    # add new subset to subset table
    res = cursor.execute('SELECT 1 FROM subset WHERE name = ?',(name,))
    if not res.fetchone():
        cursor.execute("INSERT INTO subset VALUES(?,?,?)",(None,name,desc,))
        connection.commit()
        res = cursor.execute("SELECT last_insert_rowid() FROM subset")
        subset_id = res.fetchone()[0]

        # add the substance connections to subset2substance
        for s,p in subs_dict.items():

            # set default per if none given
            if not p:
                p = default_perc
            # get substance ID
            res = cursor.execute('SELECT id FROM substance WHERE name = ?',(s,))
            subs_id = res.fetchone()
            if subs_id:
                subs_id = subs_id[0]
                # enter the entry into subset2substance
                cursor.execute('INSERT INTO subset2substance VALUES(?,?,?)',(subset_id,subs_id,p,))
                connection.commit()
            else:
                mes = f'Substance {s} not in database. Not added. Please check your input.'
                warnings.warn(mes)

    else:
        mes = f'Subset name {name} already in database. Cannot add.\nPlease choose a different name or delete the old one.'
        warnings.warn(mes,category=UserWarning)

    # close connection to database
    connection.close()


# further database curation
# -------------------------
# @CURRENTLY COMPLETELY UNTESTED, just ideas, but feel free to use 'em

# @TEST
def update_db_entry_single(table:str, column:str, new_value:Any, conditions:dict, database:str = PATH_TO_DB):
    """Update a single database entry.

    Args:
        - table (str): 
            Name of the table to update.
        - column (str): 
            Name of the Attribute to change.
        - new_value (Any): 
            New value to be set.
        - conditions (dict): 
            Further conditions.
        - database (str, optional): 
            Which database to change. Defaults to PATH_TO_DB.
    """

    # build connection to DB
    connection = sqlite3.connect(database)
    cursor = connection.cursor()

    # condition -> sql
    condition_str =  ' AND '.join([f'{x} = {y}' for x,y in conditions.items()])

    # update the entry
    cursor.execute('UPDATE ? SET ? = ? WHERE ?',(table,column,new_value,condition_str,))

    # save and close
    connection.commit()
    connection.close()


# @NOTE: this is only for adding SINGLE rows to a table WITHOUT connections
# @TEST
def enter_db_single_entry(table:str, columns:list[str], values:list[Any], database:str = PATH_TO_DB):
    """Enter a single entry into a database.

    Args:
        - table (str): 
            Which table to enter information to.
        - columns (list[str]): 
            Name of the columns to add information to.
        - values (list[Any]): 
            List of new values for the columns.
        - database (str, optional): 
            Database to add a row to. Defaults to PATH_TO_DB.
    """
    # build connection to DB
    connection = sqlite3.connect(database)
    cursor = connection.cursor()

    # add new entry to a specific table
    columns_str = '(' + ', '.join(columns) + ')'
    values_str = '(' + ', '.join(values) + ')'
    cursor.execute('INSERT INTO ? ? VALUES ?',(table,columns_str,values_str,))

    # save and close
    connection.commit()
    connection.close()

    
def generate_update_query(row: pd.Series) -> str:
    """Helper function for :py:func:`~refinegems.classes.medium.update_db_multi`. 
    Generates an update SQL query for the provided table

    Args:
        - row (pd.Series): 
            Series containing the row of a DataFrame to be used to update a table in a database
            with columns table | column | new_value | conditions

    Returns:
        str: 
            SQL query to be used to update a table in a database with the provided data
    """
    colorama_init(autoreset=True)
    table = row['table']
    conditions_dict = {k:v for k,v in [_.split('=') for _ in row["conditions"].split(';')]} # condition (str) : a=x,b=y,....
    new_value = f'\'{row["new_value"]}\'' if type(row['new_value']) == str else row['new_value']
    
    update_query = f'UPDATE {table} SET {row["column"]} = {new_value} WHERE '
    
    match table:
        
        case 'medium2substance':
            if all(_ in conditions_dict.keys() for _ in ['medium', 'substance']):
                update_query += f'''medium_id = (SELECT medium.id FROM medium WHERE medium.name = \'{conditions_dict.get("medium")}\') \
                AND substance_id = (SELECT substance.id FROM substance WHERE substance.name = \'{conditions_dict.get("substance")}\')\
                '''
            else: 
                raise ValueError(f'{Fore.MAGENTA}No medium and/or substance keys specified. Chosen table {table} cannot be updated!')
        
        case 'substance2db':
            if 'substance' in conditions_dict.keys():
                update_query += f'substance_id = (SELECT substance.id FROM substance WHERE substance.name = \'{conditions_dict.get("substance")}\')'
            else: 
                raise ValueError(f'{Fore.MAGENTA}No substance key specified. Chosen table {table} cannot be updated!')
            
        case _:
            conditions_str = ' AND '.join(row['conditions'].split(';'))
            update_query += f'{conditions_str}'
    
    return update_query

# @TEST
def generate_insert_query(row: pd.Series, cursor) -> str:
    """Helper function for :py:func:`~refinegems.classes.medium.update_db_multi`. Generate the SQL string
    for inserting a new line into the database based on a row of the table.

    Args:
        - row (pd.Series): 
            One row of the table of the parent function.

    Returns:
        str: 
            The constructed SQL string.
    """
    colorama_init(autoreset=True)
    table = row['table']
    
    # condition (str) : a=x,b=y,....
    if row['conditions'] and row['conditions'] not in ["","-"]:
        conditions_dict = {k:v for k,v in [_.split('=') for _ in row["conditions"].split(';')]}
    else: conditions_dict = None
    
    # Gather column & new_value inputs in SQL-readable format
    columns_str = f'({row["column"]})'
    value_str = row["new_value"] # Assign new_value to a variable
    if type(value_str) == str: # Check if new_value is a string
        if ',' in value_str: # Check if new_value ist actually a list
            # Get all string values as string
            value_str = ', '.join(['\'' + _.strip() + '\'' if not FLOAT_REGEX.fullmatch(_) else _ for _ in row['new_value'].split(',')])
        else: value_str = f'\'{ value_str.strip()}\'' # new_value is a string
    
    insert_query = f'INSERT INTO {row["table"]} {columns_str} VALUES '
    
    match table:
        
        case 'medium2substance':
            if not conditions_dict: 
                raise UnboundLocalError(f'{Fore.MAGENTA}No conditions column was found in the provided DataFrame. Chosen table {table} cannot be updated.')
            if all(_ in conditions_dict.keys() for _ in ['medium', 'substance']):
                insert_query += f'''(\
                    (SELECT medium.id FROM medium WHERE medium.name = \'{conditions_dict.get("medium")}\'), \
                    (SELECT substance.id FROM substance WHERE substance.name = \'{conditions_dict.get("substance")}\'), \
                    {value_str}\
                    )\
                '''
            else: 
                raise ValueError(f'{Fore.MAGENTA}No medium and/or substance keys specified. Chosen table {table} cannot be updated!')
            
        case 'subset2substance':
            if not conditions_dict: 
                raise UnboundLocalError(f'{Fore.MAGENTA}No conditions column was found in the provided DataFrame. Chosen table {table} cannot be updated.')
            if all(_ in conditions_dict.keys() for _ in ['subset', 'substance']):
                insert_query += f'''(\
                    (SELECT subset.id FROM subset WHERE subset.name = \'{conditions_dict.get("subset")}\'), \
                    (SELECT substance.id FROM substance WHERE substance.name = \'{conditions_dict.get("substance")}\'), \
                    {value_str}\
                    )\
                '''
            else: 
                raise ValueError(f'{Fore.MAGENTA}No medium and/or substance keys specified. Chosen table {table} cannot be updated!')
        
        case 'substance2db':
            if not conditions_dict: 
                raise UnboundLocalError(f'{Fore.MAGENTA}No conditions column was found in the provided DataFrame. Chosen table {table} cannot be updated.')
            if 'substance' in conditions_dict.keys():
                res = cursor.execute(f'''SELECT substance.id FROM substance WHERE substance.name = \'{conditions_dict.get("substance")}\'''')
                substance_id = res.fetchone()[0]
                ins = cursor.execute(f'''SELECT * FROM substance2db WHERE substance2db.substance_id = {substance_id} AND substance2db.db_id = \'{row["new_value"].split(",")[0].strip()}\'''').fetchone()[2].split('+')
                if len(ins) > 0:
                    cur = row['new_value'].split(',')[1].strip().split('+')
                    missing = [_ for _ in ins if _ not in cur]
                    if len(missing) > 0:
                        update_value = '+'.join(sorted(cur+missing))
                        insert_query = f'''UPDATE substance2db SET db_type = \'{update_value}\' WHERE substance2db.substance_id = {substance_id} AND szbbstance2db.db_id = \'{row["new_value"].split(",")[0].strip()}\''''
                    else:
                        return None
                # insert can be done without any problems
                else: 
                    insert_query += f'''(\
                        (SELECT substance.id FROM substance WHERE substance.name = \'{conditions_dict.get("substance")}\'), \
                        {value_str}\
                        )\
                    '''

            else: 
                raise ValueError(f'{Fore.MAGENTA}No substance key specified. Chosen table {table} cannot be updated!')
            
        case _:
            insert_query += f'({value_str})'

    return insert_query


# @TEST
def update_db_multi(data:pd.DataFrame, update_entries: bool, database:str = PATH_TO_DB):
    """Updates/Inserts multiple entries in a table from the specified database.
    Given table should have the format:
    
      row :  table | column | new_value | conditions

    Notes:

    - multiple columns and values are lists with a "," and no whitespaces
    - conditions are listed like: a=x;b=y;...

        - conditions separated by ';'
        - column and value separated by '='
        - no whitespaces

    Args:
        - data (pd.DataFrame): 
            DataFrame containing the columns table | column | new_value | conditions
        - update_entries (bool): 
            Boolean to determine whether entries should be inserted or updated. False means insert.
        - database (str, optional): 
            Path to a database. Defaults to PATH_TO_DB. 
    """
    colorama_init(autoreset=True)

    # build connection to DB
    connection = sqlite3.connect(database)
    cursor = connection.cursor()

    # iterate over the input table
    for idx,row in data.iterrows():

        # row :  table | column | new_value | conditions

        # if conditions given, update
        if row['conditions'] and row['conditions'] not in ["","-"] and update_entries: 
            query = generate_update_query(row)
        # else, insert new values
        else:
            query = generate_insert_query(row,cursor)
            if not query:
                continue   

        # update the entry
        try:
            cursor.execute(query)
        except sqlite3.IntegrityError as ie:
            print(f'{Fore.MAGENTA}{ie}')
            print(f'Ocurred with: column={row["column"]}, new_value={row["new_value"]}, condition={row["conditions"]}')
            continue

    # save and close
    connection.commit()
    connection.close()


############################################################################
# working with models
############################################################################


def medium_to_model(model:cobra.Model, medium:Medium, namespace:str='BiGG', 
                    default_flux:float=10.0, replace:bool=False, double_o2:bool=True, 
                    add:bool=True) -> Union[None, dict[str, float]]:
    """Add a medium to a COBRApy model. 

    Args:
        - model (cobra.Model): 
            A model loaded with COBRApy.
        - medium (Medium): 
            A refinegems Medium object.
        - namespace (str, optional): 
            String to set the namespace to use for the model IDs. Defaults to 'BiGG'.
        - default_flux (float, optional): 
            Set a default flux for NaN values or all. Defaults to 10.0.
        - replace (bool, optional): 
            Option to replace existing flux values with the default if set to True. Defaults to False.
        - double_o2 (bool, optional): 
            Double the oxygen amount in the medium. Defaults to True.
        - add (bool, optional): 
            If True, adds the medium to the model, else the exported medium is returned. Defaults to True.

    Returns:
        Union[None, dict[str, float]]: 
            Either none or the exported medium.
    """
    
    # export medium to cobra
    exported_medium = medium.export_to_cobra(namespace=namespace, default_flux=default_flux, replace=replace, double_o2=double_o2)

    # remove exchanges that do not exist in model
    model_exchanges = [_.id for _ in model.exchanges]
    exported_medium = {k:v for k,v in exported_medium.items() if k in model_exchanges}

    if add:
        # add to model
        model.medium = exported_medium
        return 
    else:
        return exported_medium


############################################################################
# possible entry points
############################################################################

# Function to extract SQL schema with updated SBO/media tables
def updated_db_to_schema(directory: str = '../data/database', inplace:bool=False):
    """Extracts the SQL schema from the database data.db & Transfers it into an SQL file

    Args:
        - directory(str,optional): 
            Path to the directory of the updated DB.
            Defaults to '../data/database'.
        - inplace(bool, optional): 
            If True, uses the default sql-file name, otherwise extends it with the prefix ``updated``.
    """
    
    # Not needed to be included in Schema
    NOT_TO_SCHEMA = [
        'BEGIN TRANSACTION;', 'COMMIT;', 
        'bigg_to_sbo', 'ec_to_sbo',
        'bigg_metabolites', 'bigg_reactions', 
        'modelseed_compounds'
        ]
    counter = 0 # To count rows in newly generated file

    if inplace:
        filename = 'media_db.sql'
    else:
        filename = 'updated_media_db.sql'
    
    conn = sqlite3.connect(PATH_TO_DB)
    with open(Path(directory,filename), 'w') as file:
        for line in iterdump(conn):
            if not (any(map(lambda x: x in line, NOT_TO_SCHEMA))):
                if 'CREATE TABLE' in line and counter != 0:
                    file.write(f'\n\n{line}\n')
                else: file.write(f'{line}\n')
                counter += 1
    conn.close()


# ..................................................................
# entry point for entering a medium using the command line
# @TODO since database is part of package, direct accessing possible
# @TODO more entry points and where to put them? -> cmd_access?
# @RENAME
# @TEST : is this even valid???  
# def add_medium(database:str):
    
#     # get external medium
#     medium = read_external_medium('console')

#     # add to database
#     enter_medium_into_db(medium, database)
    
#     # Generate updated SQl schema
#     updated_db_to_schema()
# ..................................................................


#!/usr/bin/env python
"""This module contains decorator functions.
"""

__author__ = "Carolin Brune"

################################################################################
# requirements
################################################################################

import warnings
from functools import wraps

################################################################################
# variables
################################################################################

################################################################################
# functions
################################################################################

# @TODO: Add note message for docs
def template(func):
    """A decorator for template functions.
    
    Template functions are not executable and are mainly for giving developers 
    a sense of how functions for the functionality this template is for should look like.
    """
    @wraps(func)
    def wrapper():
        raise RuntimeError('This function is a template for developers.\nIt cannot be executed.')
    return wrapper


def implement(func):
    """A decorator for functions that need to be implemented.
    
    Used to give a hint to other developers, that this function is not yet implemented, 
    but should be to make the program executable.
    """
    @wraps(func)
    def wrapper():
        raise NotImplementedError('The current function is just a placeholder and will be implement in the fucture.')
    
    doc_extension = """\n.. note:: \n\t*Will be coming in a future release.*"""
    if wrapper.__doc__:
        wrapper.__doc__ += doc_extension
    else:
        wrapper.__doc__ = doc_extension
    return wrapper
    

# @TODO add an option for alternative / new function
def deprecate(func):
    """A decorator to tell the user, that the function will soon be deprecated.
    
    Used to give hints to users, that their code will not work as expected, if they update 
    to a newer version.
    """
    @wraps(func)
    def wrapper(*args, **kwargs):
        mes = f'This function will be deprecated in the next major update: {func.__name__}'
        warnings.warn(mes, DeprecationWarning)
        func(*args, **kwargs)
    return wrapper


#!/usr/bin/env python
"""Helper module to work with annotations (CVTerms)

Stores dictionaries which hold information the identifiers.org syntax, has functions to add CVTerms to different entities and parse CVTerms.
"""

__author__ = "Famke Baeuerle and Gwendolyn O. Döbel"

################################################################################
# requirements
################################################################################

import cobra
import logging
from libsbml import BIOLOGICAL_QUALIFIER, BQB_IS, BQB_OCCURS_IN, BQB_IS_HOMOLOG_TO, MODEL_QUALIFIER, BQM_IS_DESCRIBED_BY, Unit, CVTerm, Species, Reaction, GeneProduct, Group, SBase

################################################################################
# variables
################################################################################

metabol_db_dict = {
                   'BIGG': 'bigg.metabolite:',
                   'BiGG': 'bigg.metabolite:',
                   'BIOCYC': 'biocyc:META:',
                   'BioCyc': 'biocyc:META:',
                   'BRENDA': 'brenda:',
                   'CHEBI': 'CHEBI:',
                   'ChEBI': 'CHEBI:',
                   'HMDB': 'hmdb:HMDB',
                   'Human Metabolome Database': 'hmdb:HMDB',
                   'INCHI': 'inchi:',
                   'InChI': 'inchi:',
                   'InChI-Key': 'inchikey:',
                   'KEGG': 'kegg.compound:',
                   #'KEGG Compound': 'kegg.compound:',
                   'METACYC': 'metacyc.compound:',
                   'MXNREF': 'metanetx.chemical:',
                   'MetaNetX': 'metanetx.chemical:',
                   'PUBCHEM': 'pubchem.compound:',
                   'REACTOME': 'reactome:',
                   'Reactome': 'reactome:',
                   'SEED': 'seed.compound:',
                   #'UPA': 'unipathway.compound:', #closed due to financial issues
                   #'UniPathway Compound': 'unipathway.compound:'
                   'VMH': 'vmhmetabolite:'
                  } #: :meta hide-value: 

reaction_db_dict = {
                    'BIGG': 'bigg.reaction:',
                    'BiGG': 'bigg.metabolite:',
                    'BioCyc': 'biocyc:META:',
                    'BRENDA': 'brenda:',
                    'EC': 'ec-code:',
                    'HMDB': 'hmdb:HMDB',
                    'KEGG': 'kegg.reaction:',
                    'METACYC': 'metacyc.reaction:',
                    'MXNREF': 'metanetx.reaction:',
                    'MetaNetX': 'metanetx.reaction:',
                    'REACTOME': 'reactome:',
                    'Reactome': 'reactome:',
                    'RHEA': 'rhea:',
                    'SEED': 'seed.reaction:',
                    #'UPA': 'unipathway.reaction:',
                    #'UniPathway Reaction': 'unipathway.reaction:'
                    'VMH': 'vmhreaction:'
                   } #: :meta hide-value: 

gene_db_dict = {
                'KEGG': 'kegg.genes:',
                'NCBI': 'ncbiprotein:',
                'REFSEQ': 'refseq:',
                'UNIPROT': 'uniprot:'
               } #: :meta hide-value: 

pathway_db_dict = {'KEGG': 'kegg.pathway:'} #: :meta hide-value: 

MIRIAM = 'https://identifiers.org/' #: :meta hide-value: 
OLD_MIRIAM = 'http://identifiers.org/' #: :meta hide-value: 

################################################################################
# functions
################################################################################

# cobra
# -----

def _add_annotations_from_dict_cobra(references:dict, entity:cobra.Reaction|cobra.Metabolite|cobra.Model) -> None:
    """Given a dictionary and a cobra object, add the former as annotations to the latter.
    The keys of the dictionary are used as the annotation labels, the values as the values.
    If the keys are already in the entity, the values will be combined (union).

    Args:
        - references (dict): 
            The dictionary with the references to add the entity.
        - entity (cobra.Reaction | cobra.Metabolite | cobra.Model): 
            The entity to add annotations to.
    """
    # add additional references from the parameter
    for db,idlist in references.items():
        if not isinstance(idlist,list):
            idlist = [idlist]
        if db in entity.annotation.keys():
            if isinstance(entity.annotation[db],list):
                entity.annotation[db] = list(set(entity.annotation[db] + idlist))
            else:
                entity.annotation[db] = list(set(list(entity.annotation[db]) + idlist))
        else:
            entity.annotation[db] = idlist



# libsbml
# -------

def add_cv_term_units(unit_id: str, unit: Unit, relation: int):
    """Adds CVTerm to a unit

    Args:
        - unit_id (str): 
            ID to add as URI to annotation
        - unit (Unit): 
            Unit to add CVTerm to
        - relation (int): 
            Provides model qualifier to be added
    """
    cv = CVTerm()
    cv.setQualifierType(MODEL_QUALIFIER)
    cv.setModelQualifierType(relation)
    
    if relation == BQM_IS_DESCRIBED_BY:
        cv.addResource(f'https://identifiers.org/{unit_id}')
    else:
        cv.addResource(f'https://identifiers.org/UO:{unit_id}')
    unit.addCVTerm(cv)


def add_cv_term_metabolites(entry: str, db_id: str, metab: Species):
    """Adds CVTerm to a metabolite

    Args:
        - entry (str): 
            Id to add as annotation
        - db_id (str): 
            Database to which entry belongs. Must be in metabol_db_dict.keys().
        - metab (Species): 
            Metabolite to add CVTerm to
    """
    if db_id == 'HMDB' or db_id == 'Human Metabolome Database':
        if entry[:4] == 'HMDB':
            entry = entry[4:]
    cv = CVTerm()
    cv.setQualifierType(BIOLOGICAL_QUALIFIER)
    cv.setBiologicalQualifierType(BQB_IS)
    cv.addResource('https://identifiers.org/' + metabol_db_dict[db_id] + entry)
    metab.addCVTerm(cv)
    metab.addCVTerm(cv)


def add_cv_term_reactions(entry: str, db_id: str, reac: Reaction):
    """Adds CVTerm to a reaction

    Args:
        - entry (str): 
            Id to add as annotation
        - db_id (str): 
            Database to which entry belongs. Must be in reaction_db_dict.keys().
        - reac (Reaction): 
            Reaction to add CVTerm to
    """
    if db_id == 'HMDB' or db_id == 'Human Metabolome Database':
        if entry[:4] == 'HMDB':
            entry = entry[4:]
    cv = CVTerm()
    cv.setQualifierType(BIOLOGICAL_QUALIFIER)
    cv.setBiologicalQualifierType(BQB_IS)
    cv.addResource(
        'https://identifiers.org/' +
        reaction_db_dict[db_id] +
        entry)
    reac.addCVTerm(cv)


def add_cv_term_genes(entry: str, db_id: str, gene: GeneProduct, lab_strain: bool=False):
    """Adds CVTerm to a gene

    Args:
        - entry (str): 
            Id to add as annotation.
        - db_id (str): 
            Database to which entry belongs. Must be in gene_db_dict.keys().
        - gene (GeneProduct): 
            Gene to add CVTerm to.
        - lab_strain (bool, optional): 
            For locally sequenced strains the qualifiers are always HOMOLOG_TO. Defaults to False.
    """
    cv = CVTerm()
    cv.setQualifierType(BIOLOGICAL_QUALIFIER)
    if lab_strain:
        cv.setBiologicalQualifierType(BQB_IS_HOMOLOG_TO)
    else:
        cv.setBiologicalQualifierType(BQB_IS)
    cv.addResource('https://identifiers.org/' + gene_db_dict[db_id] + entry)
    gene.addCVTerm(cv)


def add_cv_term_pathways(entry: str, db_id: str, path: Group):
    """Add CVTerm to a groups pathway

    Args:
        - entry (str): 
            Id to add as annotation
        - db_id (str): 
            Database to which entry belongs. Must be in pathway_db_dict.keys().
        - path (Group): 
            Pathway to add CVTerm to
    """
    cv = CVTerm()
    cv.setQualifierType(BIOLOGICAL_QUALIFIER)
    cv.setBiologicalQualifierType(BQB_IS)
    cv.addResource('https://identifiers.org/' + pathway_db_dict[db_id] + entry)
    path.addCVTerm(cv)
    

def add_cv_term_pathways_to_entity(entry: str, db_id: str, reac: Reaction):
    """Add CVTerm to a reaction as OCCURS IN pathway

    Args:
        - entry (str): 
            Id to add as annotation
        - db_id (str): 
            Database to which entry belongss
        - reac (Reaction): 
            Reaction to add CVTerm to
    """
    cv = CVTerm()
    cv.setQualifierType(BIOLOGICAL_QUALIFIER)
    cv.setBiologicalQualifierType(BQB_OCCURS_IN)
    cv.addResource('https://identifiers.org/' + pathway_db_dict[db_id] + entry)
    reac.addCVTerm(cv)


def get_id_from_cv_term(entity: SBase, db_id: str) -> list[str]:
    """Extract Id for a specific database from CVTerm

    Args:
        - entity (SBase): 
            Species, Reaction, Gene, Pathway
        - db_id (str): 
            Database of interest

    Returns:
        list[str]: 
            Ids of entity belonging to db_id
    """
    num_cvs = entity.getNumCVTerms()
    all_ids = []
    for i in range(0, num_cvs):
        ann_string = entity.getCVTerm(i)
        num_res = ann_string.getNumResources()
        ids = [ann_string.getResourceURI(r).split(
            '/')[-1] for r in range(0, num_res) if str(db_id) in ann_string.getResourceURI(r)]
        ids = [id_string.split(':')[-1] for id_string in ids if ':' in  id_string]
        all_ids.extend(ids)

    return all_ids


def generate_cvterm(qt, b_m_qt) -> CVTerm:
    """Generates a CVTerm with the provided qualifier & biological or model qualifier types

    Args:
        - qt (libSBML qualifier type): 
            BIOLOGICAL_QUALIFIER or MODEL_QUALIFIER
        - b_m_qt (libSBML qualifier): 
            BQM_IS, BQM_IS_HOMOLOG_TO, etc.

    Returns:
        CVTerm: 
            With provided qualifier & biological or model qualifier types 
    """
    cvterm = CVTerm()
    cvterm.setQualifierType(qt)
    
    if qt == BIOLOGICAL_QUALIFIER:
        cvterm.setBiologicalQualifierType(b_m_qt)
    else:
        cvterm.setModelQualifierType(b_m_qt)
    
    return cvterm 


def print_cvterm(cvterm: CVTerm):
    """Debug function: Prints the URIs contained in the provided CVTerm along with the provided qualifier & biological/model qualifier types

    Args:
        cvterm (CVTerm): 
            A libSBML CVTerm
    """
    if cvterm == None:
        logging.info('CVTerm currently empty!')
    else:
        current_b_m_qt = 0
    
        current_qt = cvterm.getQualifierType()
                
        if current_qt == BIOLOGICAL_QUALIFIER:
            current_b_m_qt = cvterm.getBiologicalQualifierType()
        elif current_qt == MODEL_QUALIFIER:
            current_b_m_qt = cvterm.getModelQualifierType()
        
        if cvterm.getNumResources() == 0:
            logging.info('No URIs present.')
        else:
            logging.info(f'Current CVTerm contains:  {cvterm.getResourceURI(0)}')
    
            for i in range(1, cvterm.getNumResources()):
                logging.info(f'                          {cvterm.getResourceURI(i)}')
                
        logging.info(f'Current CVTerm has QualifierType {current_qt} and Biological/ModelQualifierType {current_b_m_qt}.')


#!/usr/bin/env python
""" Provides functions / connections to other tools for easier access and usage.
"""

__author__ = "Famke Baeuerle und Carolin Brune"

################################################################################
# requirements
################################################################################

import cobra
import json
import memote
import pandas as pd
import shutil
import subprocess
import tempfile
import time
import warnings

from BOFdat import step1
from BOFdat import step2
from BOFdat.util import update
from BOFdat.util.update import determine_coefficients

from importlib.resources import files
from libsbml import Model as libModel
from MCC import MassChargeCuration
from pathlib import Path
from libsbml import readSBML
from sboannotator.SBOannotator import sbo_annotator
from typing import Literal

from memote.support import consistency
# needed by memote.support.consistency
from memote.support import consistency_helpers as con_helpers

from .util import test_biomass_presence
from .io import write_model_to_file

# note:
#    for BOFdat to run correctly, one needs to change 'solution.f' to 'solution.objective_value'
#    in the coenzymes_and_ions.py file of BOFdat
#    -> see forked version of it in the Draeger-lab github

################################################################################
# variables
################################################################################

################################################################################
# functions
################################################################################

# BOFdat
# ------

def adjust_BOF(genome:str, model_file:str, model:cobra.Model, dna_weight_fraction:float, weight_frac:float) -> str:
    """Adjust the model's BOF using BOFdat. Currently implemented are step 1
    DNA coefficients and step 2.

    Args:
        - genome (str): 
            Path to the genome (e.g. .fna) FASTA file.
        - model_file (str): 
            Path to the sbml (.xml) file of the model.
        - model (cobra.Model): 
            The genome-scale metabolic model (from the string above), loaded with COBRApy.
        - dna_weight_fraction (float): 
            DNA weight fraction for BOF step 1.
        - weight_frac (float): 
            Weight fraction for the second step of BOFdat (enzymes and ions)

    Returns:
        str: 
            The updated BOF reaction as a reaction string.
    """
    

    # BOFdat step 1:
    # --------------
    # dna coefficients
    dna_coefficients = step1.generate_dna_coefficients(genome,model_file,DNA_WEIGHT_FRACTION=dna_weight_fraction)
    bd_step1 = {}
    for m in dna_coefficients:
        bd_step1[m.id] = dna_coefficients[m]

    # ...........................
    # @TODO
    #    if time permits or needed, options for more coefficients can be added
    # ...........................

    # BOFdat step 2:
    # --------------
    # find inorganic ions
    selected_metabolites = step2.find_coenzymes_and_ions(model_file)
    # determine coefficients
    bd_step2 = determine_coefficients(selected_metabolites,model,weight_frac)
    bd_step2.update(bd_step1)

    # update BOF
    # ----------
    #  retrieve previously used BOF
    growth_func_list = test_biomass_presence(model)
    if len(growth_func_list) == 1: 
        objective_list = model.reactions.get_by_id(growth_func_list[0]).reaction.split(' ')
    elif len(growth_func_list) > 1:
        mes = f'Multiple BOFs found. Using {growth_func_list[0]} for BOF adjustment.'
        warnings.warn(mes,category=UserWarning)
        objective_list = model.reactions.get_by_id(growth_func_list[0]).reaction.split(' ')
    # else not needed, as if there is no BOF, a new one will be created

    # ...............................................................
    objective_reactant = {}
    objective_product = {}
    product = False
    # get reactants, product and factors from equation
    for s in objective_list:
        if s == '+':
            continue
        elif '>' in s:
            product = True
        elif '.' in s:
            factor = s
        else:
            if product:
                objective_product[s] = factor
            else:
                objective_reactant[s] = factor

    # update BOF information with data from BOFdat
    for m in bd_step2:
        if bd_step2[m] < 0:
            objective_reactant[m] = bd_step2[m]*(-1)
        else:
            objective_product[m] = bd_step2[m]

    # create new objective function
    new_objective = ' + '.join('{} {}'.format(value, key) for key, value in objective_reactant.items())
    new_objective += ' --> '
    new_objective += ' + '.join('{} {}'.format(value, key) for key, value in objective_product.items())

    return new_objective


# DIAMOND
# -------

# @ISSUE / @NOTE / @TODO
#   is this the right place to put these functions?
def run_DIAMOND_blastp(fasta:str, db:str, 
                       sensitivity:Literal['sensitive', 'more-sensitive', 'very-sensitive','ultra-sensitive']='more-sensitive',
                       coverage:float=95.0,
                       threads:int=2,
                       outdir:str=None, outname:str='DIAMOND_blastp_res.tsv') -> str:
    """Run DIAMOND in BLASTp mode.

    Args:
        - fasta (str): 
            The FASTA file to BLAST for.
        - db (str): 
            The DIAMOND database file to BLAST against
        - sensitivity (Literal['sensitive', 'more-sensitive', 'very-sensitive','ultra-sensitive'], optional): 
            Sensitivity mode for DIAMOND. 
            Defaults to 'more-sensitive'.
        - coverage (float, optional): 
            A parameter for DIAMOND
            Coverage theshold for the hits. 
            Defaults to 95.0.
        - threads (int, optional): 
            A parameter for DIAMOND.
            Number of threds to be used while BLASTing.
            Defaults to 2.
        - outdir (str, optional): 
            Path to a directory to write the output files to. 
            Defaults to None.
        - outname (str, optional): 
            Name of the result file (name only, not a path). 
            Defaults to 'DIAMOND_blastp_res.tsv'.

    Returns:
        str: 
            Path to the results of the DIAMOND BLASTp run. 
    """
    
    if outdir:
        outname = Path(outdir,'DIAMOND_blastp_res.tsv')
        logfile = Path(outdir,'log_DIAMOND_blastp.txt')
    else:
        outname = Path(outname)
        logfile = Path('log_DIAMOND_blastp.txt')
      
    # @TODO: test, if it works with different paths and their problems  
    subprocess.run([F'diamond blastp -d {db} -q {fasta} --{sensitivity} --query-cover {coverage} -p {int(threads)} -o {outname} --outfmt 6 qseqid sseqid pident length mismatch gapopen qstart qend sstart send evalue bitscore 2> {logfile}'], shell=True)

    return outname

def filter_DIAMOND_blastp_results(blasttsv:str, pid_theshold:float=90.0) -> pd.DataFrame:
    """Filter the results of a DIAMOND BLASTp run (see 
    :py:func:`~refinegems.utility.connections.run_DIAMOND_blastp`)
    by percentage identity value (PID) and extract the matching pairs of query 
    and subject IDs.

    Args:
        - blasttsv (str): 
            Path to the DIAMOND BLASTp result file.
        - pid_theshold (float, optional): 
            Threshold value for the PID. Given in percent.
            Defaults to 90.0.

    Raises:
        - ValueError: PID threshold has to be between 0.0 and 100.0

    Returns:
        pd.DataFrame: 
            A table with the columns query_ID and subject_ID containing hits from
            BLAST run with s PID higher than the given threshold value.
    """
    
    if pid_theshold > 100.0 or pid_theshold < 0.0:
        raise ValueError('PID threshold has to be between 0.0 and 100.0')
    
    # load diamond results
    diamond_results = pd.read_csv(blasttsv, sep='\t', header=None)
    diamond_results.columns = ['query_ID', 'subject_ID', 'PID', 'align_len', 'no_mismatch', 'no_gapopen', 'query_start', 'query_end', 'subject_start', 'subject_end','E-value','bitscore']
    # filter by PID
    diamond_results = diamond_results[diamond_results['PID']>=pid_theshold]
    # trim cols
    diamond_results = diamond_results[['query_ID','subject_ID']]
    
    return diamond_results


# MCC - MassChargeCuration
# ------------------------

def perform_mcc(model: cobra.Model, dir: str, apply:bool = True) -> cobra.Model:
    """Run the MassChargeCuration toll on the model and optionally directly apply 
    the solution.

    Args:
        - model (cobra.Model): 
            The model to use the tool on.
        - dir (str): 
            Path of the directory to save MCC output in.
        - apply (bool, optional): 
            If True, model is directly updated with the results. 
            Defaults to True.

    Returns:
        cobra.Model: 
            The model (updated or not)
    """

    # make temporary directory to save files for MCC in
    with tempfile.TemporaryDirectory() as temp:

        # use MCC
        if apply:
            # update model
            balancer = MassChargeCuration(model, update_ids = False, data_path=temp)
        else:
            # do not change original model
            model_copy = model.copy()
            balancer = MassChargeCuration(model_copy, update_ids = False, data_path=temp)

    # save reports
    balancer.generate_reaction_report(Path(dir,model.id+'_mcc_reactions'))
    balancer.generate_metabolite_report(Path(dir,model.id+'_mcc_metabolites'))
    balancer.generate_visual_report(Path(dir,model.id+'_mcc_visual'))

    return model



# Memote
# -------

def run_memote(model: cobra.Model, type:Literal['json','html']='html', 
               return_res:bool=False, save_res:str|None=None, verbose:bool=False) -> dict|str|None:
    """Run the memote snapshot function on a given model loaded with COBRApy.

    Args:
        - model (cobra.Model): 
            The model loaded with COBRApy.
        - type (Literal['json','html'], optional): 
            Type of report to produce. 
            Can be 'html' or 'json'. 
            Defaults to 'html'.
        - return_res (bool, optional): 
            Option to return the result. 
            Defaults to False.
        - save_res (str | None, optional): 
            If given a path string, saves the report
            under the given path. Defaults to None.
        - verbose (bool, optional): 
            Produce a more verbose ouput. 
            Defaults to False.

    Raises:
        - ValueError: Unknown input for parameter type

    Returns:
        (1) Case ``return_res = True`` and ``type = json``:
                dict: 
                    The json dictionary.

        (2) Case ``return_res = True`` and ``type = html``:
                str: 
                    The html string.

        (3) Case ``return_res = False``:
                None: 
                    no return
    """

    # verbose output I
    if verbose:
        print('\n# -------------------\n# Analyse with MEMOTE\n# -------------------')
        start = time.time()

    # run memote
    ret, res = memote.suite.api.test_model(model, sbml_version=None, results=True,
                                           pytest_args=None, exclusive=None, skip=None, 
                                           experimental=None, solver_timeout=10)
    
    # load depending on type 
    match type:
        case 'html':
            snap = memote.suite.api.snapshot_report(res, html=True)
            result = snap
        case 'json':
            snap = memote.suite.api.snapshot_report(res, html=False)
            result = json.loads(snap)
        case _:
            message = f'Unknown input for parameter type: {type} '
            raise ValueError(message)
        
    # option to save report
    if save_res:
        with open(save_res, 'w') as f:
            f.write(result)

    # verbose output II
    if verbose:
        end = time.time()
        print(F'\ttotal time: {end - start}s')

    # option to return report
    if return_res:
        return result
    

def get_memote_score(memote_report: dict) -> float:
    """Extracts MEMOTE score from report

    Args:
        - memote_report (dict): 
            Output from :py:func:`~refinegems.utility.connections.run_memote`.

    Returns:
        float: 
            MEMOTE score
    """
    return memote_report['score']['total_score']


# SBOannotator
# ------------

def run_SBOannotator(model: libModel) -> libModel:
    """Run SBOannotator on a model to annotate the SBO terms.

    Args:
        - model (libModel): 
            The model loaded with libSBML.

    Returns:
        libModel: 
            The model with corrected / added SBO terms.
    """

    dbs_scheme = files('sboannotator').joinpath('create_dbs.sql')

    with tempfile.TemporaryDirectory() as tempdir:
        write_model_to_file(model,str(Path(tempdir,'tempmodel.xml')))
        # run SBOannotator
        doc = readSBML(str(Path(tempdir,'tempmodel.xml')))
        model = doc.getModel()
        copy_scheme = shutil.copy(dbs_scheme,Path(tempdir,'dbs.sql'))
        model = sbo_annotator(doc,model,'constrained-based',str(Path(tempdir,'dbs')),str(Path(tempdir,'dud.xml')))
    return model





#!/usr/bin/env python
"""Access information from different databases or compare a model or model entities
with them. This module provides variables and function for accessing databases 
for better model curation and annotation.

The following databases have functionalities implemented:

- BiGG
- ChEBI
- KEGG
- ModelSEED
- NCBI
- UniProt
"""

__author__ = """Famke Baeuerle, Gwendolyn O. Döbel, Carolin Brune, 
            Jan-Philipp Leusch and Tobias Fehrenbach"""

############################################################################
# requirements
############################################################################

import cobra
import io
import libchebipy
import math
import numpy as np
import pandas as pd
import re
import requests
import sqlite3
import xmltodict

from Bio import Entrez, SeqIO
from Bio.KEGG import REST, Gene, Enzyme
from bioservices.kegg import KEGG
from cobra import Model as cobraModel
from multiprocessing import Pool
from ratelimit import limits, sleep_and_retry
from typing import Literal
from tqdm import tqdm

tqdm.pandas()
pd.options.mode.chained_assignment = None # suppresses the pandas SettingWithCopyWarning; comment out before developing!!

from .connections import run_DIAMOND_blastp, filter_DIAMOND_blastp_results
from .databases import PATH_TO_DB
from .io import load_a_table_from_database, create_missing_genes_protein_fasta
from .util import VALID_COMPARTMENTS

############################################################################
# variables
############################################################################

# database urls
# -------------

BIGG_REACTIONS_URL = 'http://bigg.ucsd.edu/api/v2/universal/reactions/' #: :meta: 
BIGG_METABOLITES_URL = 'http://bigg.ucsd.edu/api/v2/universal/metabolites/' #: :meta: 

# Compartments in BiGG namespace
# ------------------------------

ALL_BIGG_COMPARTMENTS_ONE_LETTER = ('c', 'e', 'p', 'm', 'x', 'r', 'v', 'n', 'g', 'u', 'l', 'h', 'f', 's', 'i', 'w', 'y') #: :meta: 
ALL_BIGG_COMPARTMENTS_TWO_LETTER = ('im', 'cx', 'um', 'cm', 'mm') #: :meta: 

############################################################################
# functions
############################################################################

# BiGG
# ----
"""Map an ID to information in BiGG or compare model entities to BiGG.
"""

# @NOTE: Add to reaction handling for `build_reac_bigg`?
def add_stoichiometric_values_to_reacs_from_bigg(missing_reacs: pd.DataFrame) -> pd.DataFrame:
    """Adds for each reaction a dictionary containing the reactants & products as dictionaries with the BiGG Metabolite 
    ID as key & the respective absolute stoichiometric value as value
        
    Args:
        - missing_reacs (pd.DataFrame): 
            Table containing missing reactions (Only requires a column containing BiGG IDs)
            
    Returns:
        pd.DataFrame: 
            Table where for each BiGG reaction ID a dictionary containing reactants & products exists 
    """
    
    def get_reactants_and_products_dicts(reaction_id: str) -> list[dict]:
        reactants = {}
        products = {}
        
        metabs_from_reac = requests.get(BIGG_REACTIONS_URL + reaction_id).json()['metabolites']

        for compound_dict in metabs_from_reac:
            complete_bigg_id = None
            if compound_dict.get('compartment_bigg_id'):
                complete_bigg_id = f"{compound_dict.get('bigg_id')}_{compound_dict.get('compartment_bigg_id')}"
            else:
                complete_bigg_id = compound_dict.get('bigg_id')
            if compound_dict.get('stoichiometry') < 0:
                reactants[complete_bigg_id] = abs(compound_dict.get('stoichiometry'))
            elif compound_dict.get('stoichiometry') > 0:
                products[complete_bigg_id] = abs(compound_dict.get('stoichiometry'))
                
        return str({'reactants': reactants, 'products': products})
                
    missing_reacs['bigg_reaction']= missing_reacs.apply(
        lambda row: get_reactants_and_products_dicts(str(row['bigg_id'])), axis=1)  #, missing_reacs['bigg_products'], result_type='expand'
      
    return missing_reacs


# @NOTE: Add to reaction handling for `build_reac_bigg`?
@sleep_and_retry
@limits(calls=10, period=1)
def get_reaction_compartment_from_bigg(bigg_id: str) -> str:
    """Retrieves the compatment(s) a reaction is hapening in from the BiGG reaction identifier via the metabolites

    Args:
        - bigg_id (str): 
            BiGG reaction identifier

    Returns:

        (1) Case ``compartment in VALID_COMPARTMENTS.keys()``
                str: 
                    Either 
                    
                    - Compartment of the provided reaction if reaction in single compartment
                    - 'exchange' if reaction in multiple compartments

        (2) Case not a valid compartment:
                np.nan: 
                    'NaN' if one of the found compartments is not in :py:const:`~refinegems.utility.util.VALID_COMPARTMENTS`
    """
    
    metabs_from_reac = requests.get(BIGG_REACTIONS_URL + bigg_id, allow_redirects=False).json()['metabolites']
            
    comps = [comp_dict.get('compartment_bigg_id') for comp_dict in metabs_from_reac]  # Get all compartments for reaction
    contained_in_compartments = [(comp in VALID_COMPARTMENTS.keys()) for comp in comps]  # Get True for correct compartment        
    if not all(contained_in_compartments):  # At least one compartment not correct
        return np.nan
    else:  # All compartments correct
        if len(set(comps)) == 1:  # Set of found compartments of reaction = 1: Reaction happens in one compartment
            return comps[0]
        else:  # Not so important but do not remove reaction as reaction in correct compartments
            return 'exchange'  # Probably exchange reaction

def get_BiGG_metabs_annot_via_dbid(metabolite:cobra.Metabolite, 
                                        id:str, dbcol:str, 
                                        compartment:str = 'c') -> None:
    """Search for a BiGG ID and add it to a metabolite annotation. 
    The search is based on a column name of the BiGG metabolite table
    and an ID to search for. Additionally, using the given compartment name, 
    the found IDs are filtered for matching compartments.

    Args:
        - metabolite (cobra.Metabolite): 
            The metabolite. Needs to a a COBRApy Metabolte object.
        - id (str): 
            The ID to search for in the database.
        - dbcol (str): 
            Name of the column of the database to check the ID against.
        - compartment (str, optional): 
            The compartment name. Needs to be a valid BiGG compartment ID. 
            Defaults to 'c'.
    """
    # check, if a BiGG annotation is given
    if not 'bigg.metabolite' in metabolite.annotation.keys():
        # seach the database for the found id
        bigg_search = load_a_table_from_database(
            f'SELECT * FROM bigg_metabolites WHERE \'{dbcol}\' = \'{id}\'',
            query=True)
        if len(bigg_search) > 0:
            # check, if the matches also match the compartment
            metabolite.annotation['bigg.metabolite'] = [_.rsplit('_',1)[0] for _ in bigg_search['id'].tolist() if _.endswith(f'_{compartment}')]
            # add final matches to the annotations of the metabolite
            if len(metabolite.annotation['bigg.metabolite']) == 0:
                metabolite.annotation.pop('bigg.metabolite')
 
 
def add_annotations_from_BiGG_metabs(metabolite:cobra.Metabolite) -> None:
    """Check a cobra.metabolite for bigg.metabolite annotations. If they exists, 
    search for more annotations in the BiGG database and add them to the metabolite.

    Args:
        - metabolite (cobra.Metabolite): 
            The metabolite object.
    """
    if 'bigg.metabolite' in metabolite.annotation.keys():
        bigg_information = load_a_table_from_database(
            'SELECT * FROM bigg_metabolites WHERE id = \'' + f'\' OR id = \''.join(metabolite.annotation['bigg.metabolite']) + '\'',
            query=True)
        db_id_bigg = {'BioCyc':'biocyc', 'MetaNetX (MNX) Chemical':'metanetx.chemical','SEED Compound':'seed.compound','CHEBI':'chebi', 'KEGG Compound':'kegg.compound'}
        for db in db_id_bigg:
            info = list(set(bigg_information[db].dropna().to_list()))
            if len(info) > 0:
                info = ','.join(info)
                info = [x.strip() for x in info.split(',')] # make sure all entries are a separate list object
                if db_id_bigg[db] in metabolite.annotation.keys():
                    metabolite.annotation[db_id_bigg[db]] = list(set(info + metabolite.annotation[db_id_bigg[db]]))
                else:
                    metabolite.annotation[db_id_bigg[db]] = info


def _add_annotations_from_bigg_reac_row(row:pd.Series, reac:cobra.Reaction) -> None:
    """Given a row of the BiGG reaction database table and a cobra.Reaction object,
    extend the annotation of the latter with the information of the former.

    Args:
        - row (pd.Series): 
            The row of the database table.
        - reac (cobra.Reaction): 
            The reaction object.
    """
    
    dbnames = {'RHEA':'rhea','BioCyc':'biocyc','MetaNetX (MNX) Equation':'metanetx.reaction','EC Number':'ec-code'}
    for dbname,dbprefix in dbnames.items():
        if row[dbname]:
            ids_to_add = row[dbname].split(',')
            if dbprefix in reac.annotation.keys():
                reac.annotation[dbprefix] = list(set(reac.annotation[dbprefix]).union(set(ids_to_add)))
            else:
                reac.annotation[dbprefix] = ids_to_add


# @TEST
# @NOTE : A lot of warnings
def keep_only_bigg_reactions_in_certain_compartments(complete_df: pd.DataFrame) -> pd.DataFrame:
    """Extracts all possible BiGG ID variations from database for a BiGG reaction ID, gets the metabolite compartments 
    & returns table containing only reactions which happen in one of the provided compartments
        
    Args:
        - complete_df (pd.DataFrame): 
            Table containing at least the column 'bigg_id'.
        
    Returns:
        pd.DataFrame: 
            Table containing reactions & their compartments
    """
    tqdm.pandas()
    
    # (1) Find all occurrencs of a BiGG reaction ID in bigg_reactions table in database
    def get_all_similar_bigg_ids(bigg_id_in: str) -> list[str]:
        
        if '_' in bigg_id_in: bigg_id = re.split('_([a-zA-Z]|[0-9])$', bigg_id_in)[0]
        elif bigg_id_in.endswith(ALL_BIGG_COMPARTMENTS_ONE_LETTER): bigg_id = bigg_id_in[:-1]
        elif bigg_id_in.endswith(ALL_BIGG_COMPARTMENTS_TWO_LETTER): bigg_id = bigg_id_in[:-2]
        else: bigg_id = bigg_id_in
        
        query = f"SELECT id, INSTR(id, '{bigg_id}') bi FROM bigg_reactions WHERE bi > 0"
        result = con.execute(query).fetchall()
        result = [result_tuple[0] for result_tuple in result] if result else [bigg_id_in]
        result = [res for res in result if compare_bigg_ids(bigg_id, res)]
        return result
    
    # (2) Use list of all BiGG IDs obtained from database table bigg_reactions to get 'metabolites'
    # get_react_compartment moved to outer scope due to multiprocessing Pool (see line 95)
    def multi_get_reaction_compartment(complete_df: pd.DataFrame) -> list:
        """Takes a dataframe and runs get_reaction_compartment() in multiple
        processes on the 'bigg_id' column

        Args:
            - complete_df (pd.DataFrame): Table containing at least the columns 'bigg_id' & 'KEGG'/'BioCyc'

        Returns:
            list: List of compartments
        """
        with Pool() as pool:
            results = []
            for out in tqdm(pool.imap(get_reaction_compartment_from_bigg, complete_df.loc[:, "bigg_id"], chunksize=20), total=len(complete_df)):
                results.append(out)

        return results

    print(complete_df.columns)
    
    # Connect to database & get similar IDs (1)
    print('Getting all similar IDs...')
    con = sqlite3.connect(PATH_TO_DB)  # Open connection to database
    complete_df.loc[:, 'bigg_id_list'] = complete_df.loc[:, 'bigg_id'].progress_map(get_all_similar_bigg_ids)
    # complete_df.progress_apply(get_all_similar_bigg_ids, axis=1)
    con.close()  # Close connection to database

    # Adjust table to contain one BiGG ID per row from bigg_id_list (1)
    complete_df.loc[:, 'id_group'] = complete_df['bigg_id'].ne(complete_df['bigg_id'].shift()).cumsum()  # Group similar IDs
    complete_df.drop(labels='bigg_id', axis=1, inplace=True)  # Drop 'bigg_id' as no longer required
    complete_df = complete_df.explode('bigg_id_list', ignore_index=True)  # Expand 'bigg_id_list' column
    complete_df.rename(columns={'bigg_id_list': 'bigg_id'}, inplace=True)  # Rename 'bigg_id_list' to 'bigg_id'

    # (2) Get all compartments for each reaction from BiGG database API
    print(f'Getting all IDs with correct compartment {VALID_COMPARTMENTS.keys()}...')
    results = multi_get_reaction_compartment(complete_df)
    complete_df["compartment"] = results

    # complete_df.progress_apply(get_reaction_compartment, axis=1)  # (2)

    # (3) Remove reactions with compartment = NaN
    complete_df.dropna(subset=['compartment'], inplace=True)

    return complete_df




# @DISCUSSION
# -----------
def compare_bigg_ids(id1: str, id2: str) -> bool:
    """Compares two BiGG strings/IDs & Returns True if one BiGG ID matches most of the other

    Args:
        - id1 (str): 
            ID 1
        - id2 (str): 
            ID 2

    Returns:
        bool: 
            Indicates if most of one string contained in the other
    """
    id1_split, id2_split, id1_single_comp, id2_single_comp, id1_comp, id2_comp = None, None, None, None, None, None
    
    if '_' in id1: id1_split = re.split('_([a-zA-Z]|[0-9])$', id1)[0]
    if '_' in id2: id2_split = re.split('_([a-zA-Z]|[0-9])$', id2)[0]
    if id1.endswith(ALL_BIGG_COMPARTMENTS_ONE_LETTER): id1_single_comp = id1[:-1]
    if id2.endswith(ALL_BIGG_COMPARTMENTS_ONE_LETTER): id2_single_comp = id2[:-1]
    if id1.endswith(ALL_BIGG_COMPARTMENTS_TWO_LETTER): id1_comp = id1[:-2]
    if id2.endswith(ALL_BIGG_COMPARTMENTS_TWO_LETTER): id2_comp = id2[:-2]
    
    similar_ids = False
    if id1 == id2: similar_ids = True  # Both IDs are same
    
    elif id1_split and id2_split and (id1_split == id2_split): similar_ids = True # Both IDs are same but from different compartments
    elif id2_split and (id1 == id2_split): similar_ids = True # - "" -
    elif id1_split and (id1_split == id2): similar_ids = True # - "" -
    
    elif id1_single_comp and id2_single_comp and (id1_single_comp == id2_single_comp): similar_ids = True
    elif id2_single_comp and (id1 == id2_single_comp): similar_ids = True
    elif id1_single_comp and (id1_single_comp == id2_single_comp): similar_ids = True 
    
    elif id1_comp and id2_comp and (id1_comp == id2_comp): similar_ids = True
    elif id2_comp and (id1 == id2_comp): similar_ids = True
    elif id1_comp and (id1_comp == id2): similar_ids = True
    
    elif id1_split and id2_single_comp and (id1_split == id2_single_comp): similar_ids = True
    elif id2_split and id1_single_comp and (id1_single_comp == id2_split): similar_ids = True
    
    elif id1_split and id2_comp and (id1_split == id2_comp): similar_ids = True
    elif id2_split and id1_comp and (id1_comp == id2_split): similar_ids = True
    
    elif id1_comp and id2_single_comp and (id1_comp == id2_single_comp): similar_ids = True
    elif id2_comp and id1_single_comp and (id1_single_comp == id2_comp): similar_ids = True
    
    else: similar_ids = False

    return similar_ids


# @TEST
def get_bigg_db_mapping(map_to:str='BioCyc', metabolites:bool=True) -> pd.DataFrame:
    """Download a mapping of BiGG IDs to a specified database.

    Args:
        map_to (str, optional): 
            Name of the database to map to. 
            Ideally a column of the table in the database, 
            but SEED, KEGG and BioCyc are valid as well. 
            Defaults to 'BioCyc'.
        metabolites (bool, optional): 
            Flag to map reaction (False) or metabolite (True) IDs. 
            Defaults to True.

    Raises:
        - KeyError: Given database name not found in database. Cannot perform mapping.

    Returns:
        pd.DataFrame: 
            The mapping as a table.
    """

    # adjust name to map to if necessary
    reac_or_comp = 'Compound' if metabolites else 'Reaction'
    table_name = 'bigg_metabolites' if metabolites else 'bigg_reactions'
    if map_to in ['SEED','KEGG','Reactome']:
        map_to =  ' '.join([map_to, reac_or_comp])
    
    # download BiGG tables from database
    # ----------------------------------
    # build connection to DB
    connection = sqlite3.connect(PATH_TO_DB)
    cursor = connection.cursor()

    # retrieve only mappings to a specific database
    result = cursor.execute('SELECT 1 FROM PRAGMA_TABLE_INFO(?) WHERE name = ?',(table_name,map_to))
    possible_db = result.fetchone()
    if possible_db:
        query = f'SELECT * FROM {table_name} WHERE {map_to} IS NOT NULL'
    else:
        raise KeyError('Given database name not found in database. Cannot perform mapping.')

    # actually load data
    data = load_a_table_from_database(query)
    data = data.explode(map_to, ignore_index=True)

    # reduce columns to mapping only
    data = data[['id',map_to]]
    data.rename(columns={'id':'bigg_id'}, inplace=True)

    # filter for compartment in case of reactions
    if not metabolites:
        data = keep_only_bigg_reactions_in_certain_compartments(data)

    # close connection to database
    connection.close()
    
    return data


# ChEBI
# -----
"""Primarily used for curating the media database. Obtain information about 
metabolites (table format)."""

# @DISCUSSION: Add to metabolite handling for `.entities.build_metabolite_bigg`?
def add_charges_chemical_formulae_to_metabs(missing_metabs: pd.DataFrame) -> pd.DataFrame:
   """Adds charges & chemical formulae from CHEBI/BiGG to the provided dataframe

   Args:
      - missing_metabs (pd.DataFrame): 
         Table containing metabolites & the respective CHEBI & BiGG IDs
         
   Returns:
      pd.DataFrame: 
         Input table extended with the charges & chemical formulas obtained from CHEBI/BiGG
   """
   
   # Finds the charges through the ChEBI/BiGG API, defaults to: 0
   def find_charge(row: pd.Series) -> int:
      chebi_id = str(int(row.get('ChEBI'))) if not math.isnan(float(row.get('ChEBI'))) else None
      bigg_id = str(row.get('bigg_id'))
      charge = None
      if chebi_id:  # Get charge from ChEBI (Returns always a charge)
         chebi_entity = libchebipy.ChebiEntity('CHEBI:' + chebi_id)
         return chebi_entity.get_charge()
      elif bigg_id != 'nan':  # Get charge from BiGG if no ChEBI ID available
         try:
            charge = requests.get(BIGG_METABOLITES_URL + bigg_id[:-2]).json()['charges'][0]  # Take first charge
         except ValueError:
            pass   
         # If no charge was found, charge=0
         return charge if charge else 0
   
   # Finds the chemical formula through the ChEBI/BiGG API, defaults to: 'No formula'
   def find_formula(row: pd.Series) -> str:
      chebi_id = str(int(row.get('ChEBI'))) if not math.isnan(float(row.get('ChEBI'))) else None
      bigg_id, chem_form = str(row.get('bigg_id')), str(row.get('Chemical Formula'))
      chem_formula = None
      if chebi_id: # Get formula from ChEBI
         chebi_entity = libchebipy.ChebiEntity('CHEBI:' + chebi_id)
         chem_formula = chebi_entity.get_formula()
      if not chem_formula:  # If no formula was found with ChEBI/No ChEBI ID available
         if bigg_id != 'nan': # Get formula from BiGG
            try:
               chem_formula = requests.get(BIGG_METABOLITES_URL + bigg_id[:-2]).json()['formulae'][0]  # Take first formula
            except (ValueError, IndexError) as e:
               pass
         if not chem_formula: # If no formula was found with BiGG ID
            # Get formula already existing in dataframe or set to 'No formula'
            chem_formula = chem_form if chem_form != 'nan' else '*'
      return chem_formula
   
   missing_metabs['charge'] = missing_metabs.apply(find_charge, axis=1)
   missing_metabs['New Chemical Formula'] = missing_metabs.apply(find_formula, axis=1)
   missing_metabs['Chemical Formula'] = missing_metabs['New Chemical Formula']
   missing_metabs.drop('New Chemical Formula', axis=1, inplace=True)
   
   return missing_metabs

def add_info_from_ChEBI_BiGG(missing_metabs: pd.DataFrame, charge=True, formula=True, iupac=True) -> pd.DataFrame:
   """Adds information from CHEBI/BiGG to the provided dataframe.

   The following informations can be added:

   - charge
   - formula
   - iupac (name)

   Args:
      - missing_metabs (pd.DataFrame): 
         Table containing metabolites & the respective ChEBI & BiGG IDs
         
   Returns:
      pd.DataFrame: 
         Input table extended with the charges & chemical formulas obtained from ChEBI/BiGG.
   """

   # check if a row contains a ChEBI ID, take the first and make sure its in the format: CHEBI:234567
   def get_chebi_id(row: pd.Series) -> str:

      chebi = row.get('ChEBI')
      if pd.isnull(chebi):
         return None
      elif type(chebi) == str: # check for mulitple entries (assuming first one is the best fitting one)
         chebi = chebi.split(',')[0]
         if 'CHEBI' in chebi:
            return chebi
         else:
            return 'CHEBI:' + chebi
      else:
         return 'CHEBI:' + str(chebi)

   # Finds the charges through the ChEBI/BiGG API, defaults to: 0
   def find_charge(row: pd.Series) -> int:
      chebi_id = get_chebi_id(row)
      bigg_id = str(row.get('bigg_id'))
      charge = None
      if chebi_id:  # Get charge from ChEBI (Returns always a charge)
         chebi_entity = libchebipy.ChebiEntity(chebi_id)
         return chebi_entity.get_charge()
      elif bigg_id != 'nan':  # Get charge from BiGG if no ChEBI ID available
         try:
            charge = requests.get(BIGG_METABOLITES_URL + bigg_id[:-2]).json()['charges'][0]  # Take first charge
         except ValueError:
            pass   
         # If no charge was found, charge=0
         return charge if charge else 0
    
    # Finds the chemical formula through the ChEBI/BiGG API, defaults to: 'No formula'
   def find_formula(row: pd.Series) -> str:
      chebi_id = get_chebi_id(row)
      bigg_id, chem_form = str(row.get('bigg_id')), str(row.get('Chemical Formula'))
      chem_formula = None
      if chebi_id: # Get formula from ChEBI
         chebi_entity = libchebipy.ChebiEntity(chebi_id)
         chem_formula = chebi_entity.get_formula()
      if not chem_formula:  # If no formula was found with ChEBI/No ChEBI ID available
         if bigg_id != 'nan': # Get formula from BiGG
            try:
               chem_formula = requests.get(BIGG_METABOLITES_URL + bigg_id[:-2]).json()['formulae'][0]  # Take first formula
            except ValueError:
               pass
         if not chem_formula: # If no formula was found with BiGG ID
            # Get formula already existing in dataframe or set to 'No formula'
            chem_formula = chem_form if chem_form != 'nan' else 'No formula'
      return chem_formula
   
   # using the ChEBI ID, retrieve the IUPAC name from the ChEBI database.
   def find_iupac(row: pd.Series) -> str:

      chebi_id = get_chebi_id(row)
      if chebi_id:
         chebi_entity = libchebipy.ChebiEntity(chebi_id)
         # only take the IUPAC names
         iupac_names = []
         for name in chebi_entity.get_names():
            print(name)
            if name.get_source() == 'IUPAC':
               iupac_names.append(name.get_name())
               break

         if len(iupac_names) > 0:
            iupac_name = ', '.join(iupac_names)
         else:
            iupac_name = None
      else:
         iupac_name = None

      return iupac_name

   if charge:
      missing_metabs['charge'] = missing_metabs.apply(find_charge, axis=1)
   if formula:
      missing_metabs['New Chemical Formula'] = missing_metabs.apply(find_formula, axis=1)
      missing_metabs['Chemical Formula'] = missing_metabs['New Chemical Formula']
      missing_metabs.drop('New Chemical Formula', axis=1, inplace=True)
   if iupac:
      missing_metabs['ChEBI_IUPAC'] = missing_metabs.apply(find_iupac, axis=1)
   
   return missing_metabs


# KEGG
# ----
"""Retrieve and parse information from the KEGG database. 
"""

def get_kegg_genes(organismid: str) -> pd.DataFrame:
    """Extracts list of genes from KEGG given an organism

    Args:
        - organismid (str): 
            KEGG ID of organism which the model is based on

    Returns:
        pd.DataFrame: 
            Table of all genes denoted in KEGG for the organism
    """
    k = KEGG()
    gene_list = k.list(organismid)

    return pd.read_table(io.StringIO(gene_list), header=None)


def parse_KEGG_gene(locus_tag:str) -> dict:
    """Based on a locus tag, fetch the corresponding KEGG entry and
    parse it into a dictionary containing the following information (if available):
    
    - ec-code
    - orthology
    - references

    Args:
        - locus_tag (str): 
            The locus in the format <organism_id>:<locus_tag>

    Returns:
        dict: 
            The collected information.
    """
    
    gene_info = dict()
    gene_info['orgid:locus'] = locus_tag
    
    # retireve KEGG gene entry 
    try: 
        gene_entry = list(Gene.parse(REST.kegg_get(locus_tag)))[0]
    except Exception as e:
        # @TODO : warning / logging
        gene_entry = None
    
    # skip, if no entry found
    if not gene_entry:
        gene_info['ec-code'] = None
        return gene_info
    
    # extract orthology and ec-code
    if len(gene_entry.orthology) > 0:
        # get KEGG orthology ID
        kegg_orthology = [_[0] for _ in gene_entry.orthology]
        gene_info['kegg.orthology'] = kegg_orthology
        # get EC number
        ec_numbers = [re.search('(?<=EC:).*(?=\])',orth[1]).group(0) for orth in gene_entry.orthology if re.search('(?<=EC:).*(?=\])',orth[1])]
        if isinstance(ec_numbers,list) and len(ec_numbers) > 0:
            gene_info['ec-code'] = [ec for ec_str in ec_numbers for ec in ec_str.split(' ')]
            
    if not 'ec-code' in gene_info.keys():
        gene_info['ec-code'] = None
        
    # get more information about connections to other databases
    if len(gene_entry.dblinks) > 0:
        for dbname,ids in gene_entry.dblinks:
            conform_dbname = re.sub(pattern='(NCBI)(.*)(ID$)', repl='\\1\\2',string=dbname) # Remove ID if NCBI in name
            conform_dbname = re.sub('[^\w]','',conform_dbname) # remove special signs except underscore
            conform_dbname = conform_dbname.lower() # make lower case
            gene_info[conform_dbname] = ids
            
    return gene_info
            
    
def parse_KEGG_ec(ec:str) -> dict:
    """Based on an EC number, fetch the corresponding KEGG entry and
    parse it into a dictionary containing the following information (if available):
    
    - ec-code
    - id (kegg.reference)
    - equation
    - reference 
    - pathway

    Args:
        - ec (str): 
            The EC number in the format 'x.x.x.x'

    Returns:
        dict: 
            The collected information about the KEGG entry.
    """
    
    ec_info = dict()
    ec_info['ec-code'] = ec
    
    # retrieve KEGG entry
    # @TODO : add time restraint and tries for time out only
    try:
        ec_entry = list(Enzyme.parse(REST.kegg_get(ec)))[0]
    except Exception as e:
        # @TODO logging / warning
        ec_entry = None
        ec_info['id'] = None
        ec_info['equation'] = None
        ec_info['reference'] = None
        return ec_info
    
    # retrieve reaction information from entry 
    rn_numbers = [re.search('(?<=RN:).*(?=\])',reac).group(0) for reac in ec_entry.reaction if re.search('(?<=RN:).*(?=\])',reac)]
    if len(rn_numbers) > 0:
        ec_info['id'] = [_.split(' ') for _ in rn_numbers]
        ec_info['equation'] = None
    else:
        ec_info['id'] = None
        ec_info['equation'] = ec_entry.reaction
        
    # retrieve database links from entry
    refs = dict()
    # orthology not possible with biopython
    if len(ec_entry.pathway) > 0:
        refs['kegg.pathway'] = [_[1] for _ in ec_entry.pathway]
    # @TODO extend as needed
    if len(ec_entry.dblinks) > 0:
        for dbname, ids in ec_entry.dblinks:
            if 'BRENDA' in dbname:
                refs['brenda'] = ids
            if 'CAS' == dbname:
                refs['cas'] = ids
    ec_info['reference'] = refs
    
    return ec_info


# @TODO open issues
def kegg_reaction_parser(rn_id:str) -> dict: 
    """Get the entry of a KEGG reaction ID and 
    parse the information into a dictionary.

    Args:
        - rn_id (str): 
            A reaction ID existing in KEGG.

    Returns:
        dict: 
            The KEGG entry information as a dictionary.
    """

    # get KEGG reaction entry
    try:
        kegg_reac = REST.kegg_get(F'rn:{rn_id}')
        kegg_reac = kegg_reac.read()
    except Exception as e:
        # @TODO
        return None

    # parse the entry for necessary information
    features = {}
    db_entries = []
    pathways = []
    rc = []
    references = {'kegg.reaction':rn_id}
    collect = False
    for line in kegg_reac.split('\n'):
        if line:
            if line.startswith('NAME'):
                features['name'] = line.replace('NAME','',1).strip()
            elif line.startswith('EQUATION'):
                features['equation'] = line.replace('EQUATION','',1).strip()
            elif line.startswith('ENZYME'):
                references['ec-code'] = line.replace('ENZYME','',1).strip()
            elif line.startswith('RCLASS'):
                rc.append(line.replace('RCLASS','',1).strip().split(' ')[0])
                collect = True
            elif line.startswith('PATHWAY'):
                pathways.append(line.replace('PATHWAY','',1).strip().split(' ')[0])
                collect = True
            elif line.startswith('DBLINKS'):
                db_entries.append(line.replace('DBLINKS','',1).strip())
                collect = True
            elif collect == True and line[0] != '/':
                if len(db_entries) == 0:
                    if line[0].isupper():
                        collect = False
                    else:
                        line = line.strip()
                        if line.startswith('RC'):
                            rc.append(line.split(' ')[0])
                        else:
                            pathways.append(line.split(' ')[0])
                else:
                    db_entries.append(line.strip())
            else:
                continue

    # parse references
    for entry in db_entries:
        db, identifier = entry.split(':')
        db = db.strip().lower()
        if db in references:
            references[db] = references[db].append(identifier)
        else:
            references[db] = [identifier.strip()]
    if len(pathways) > 0:
        references['kegg.pathway'] = pathways
    if len(rc) > 0:
        references['kegg.rclass'] = rc
    if len(references) > 0:
        features['db'] = references

    return features


# ModelSEED
# ---------
""" Reports mismatches in charges and formulae based on ModelSEED

Extracts ModelSEED data from a given tsv file, extracts all metabolites from a given model. Both lists of metabolites are compared by charge and formula.
"""

def get_modelseed_compounds() -> pd.DataFrame:
    """Extracts compounds from ModelSEED which have BiGG Ids

    Returns:
        pd.DataFrame: 
            Table containing ModelSEED data
    """
    # Get only rows where BiGG is contained
    com = load_a_table_from_database("SELECT *, INSTR(aliases, 'BiGG:') bigg FROM modelseed_compounds WHERE bigg > 0")

    def get_bigg_ids(aliases):
        try:
            aliases_list = aliases.split('|')
            bigg = [x[6:] for x in aliases_list if re.search('BiGG: .*', x)]
            return bigg[0]
        except (IndexError, AttributeError):
            return None

    com['BiGG'] = com.apply(lambda row: get_bigg_ids(row['aliases']), axis=1)

    return com.loc[:, ['id', 'name', 'formula', 'mass', 'charge', 'BiGG']]


def get_model_charges(model: cobraModel) -> pd.DataFrame:
    """Extracts all metabolites from model

    Args:
        - model (cobraModel): 
            Model loaded with COBRApy

    Returns:
        pd.DataFrame: 
            Table containing charges and formulae of model metabolites
    """
    charges = {}
    for metab in model.metabolites:
        charges[metab.id[:-2]] = [metab.charge, metab.formula]

    df_charges = pd.DataFrame.from_dict(
        charges,
        orient='index',
        columns=[
            'charge_model',
            'formula_model']).reset_index().rename(
        columns={
            'index': 'BiGG'})

    return df_charges


def get_modelseed_charges(modelseed_compounds: pd.DataFrame) -> pd.DataFrame:
    """Extract table with BiGG, charges and formulae

    Args:
        - modelseed_compounds (pd.DataFrame): 
            ModelSEED data. Output from :py:func:`~refinegems.utility.db_access.get_modelseed_compounds`.

    Returns:
        pd.DataFrame: 
            Table containing charges and formulae of ModelSEED metabolites
    """
    modelseed_compounds = modelseed_compounds.loc[:, ['charge', 'BiGG', 'formula']].rename(
        columns={'charge': 'charge_modelseed', 'formula': 'formula_modelseed'})
    lst_col = 'BiGG'
    x = modelseed_compounds.assign(
        **{lst_col: modelseed_compounds[lst_col].str.split(';')})
    df_ms = pd.DataFrame({col: np.repeat(x[col].values, x[lst_col].str.len()) for col in x.columns.difference(
        [lst_col])}).assign(**{lst_col: np.concatenate(x[lst_col].values)})[x.columns.tolist()]
    return df_ms


def compare_model_modelseed(model_charges: pd.DataFrame, modelseed_charges: pd.DataFrame) -> pd.DataFrame:
    """Compares tables with charges / formulae from model & modelseed

    Args:
        - model_charges (pd.DataFrame): 
            Charges and formulae of model metabolites. Output of :py:func:`~refinegems.utility.db_access.get_model_charges`.
        - modelseed_charges (pd.DataFrame): 
            Charges and formulae of ModelSEED metabolites. Output of :py:func:`~refinegems.utility.db_access.get_modelseed_charges`.

    Returns:
        pd.DataFrame: 
            Table containing info whether charges / formulae match
    """
    df_comp = pd.merge(model_charges, modelseed_charges, on='BiGG', how='left')

    def f(x):
        return True if float(
            x['charge_model']) == x['charge_modelseed'] else False

    def g(x):
        return True if x['formula_model'] == x['formula_modelseed'] else False

    df_comp['charge_match'] = df_comp.apply(f, axis=1)
    df_comp['formula_match'] = df_comp.apply(g, axis=1)

    return df_comp


def get_charge_mismatch(df_comp: pd.DataFrame) -> pd.DataFrame:
    """Extracts metabolites with charge mismatch of model & modelseed

    Args:
        df_comp (pd.DataFrame): 
            Charge and formula mismatches. Output from :py:func:`~refinegems.utility.db_access.compare_model_modelseed`.

    Returns:
        pd.DataFrame: 
            Table containing metabolites with charge mismatch
    """
    return df_comp.loc[~df_comp['charge_match']].dropna(
        subset=['charge_modelseed'])


def get_formula_mismatch(df_comp: pd.DataFrame) -> pd.DataFrame:
    """Extracts metabolites with formula mismatch of model & modelseed

    Args:
        df_comp (pd.DataFrame): 
            Charge and formula mismatches. Output from :py:func:`~refinegems.utility.db_access.compare_model_modelseed`.

    Returns:
        pd.DataFrame: 
            Table containing metabolites with formula mismatch
    """
    return df_comp.loc[~df_comp['formula_match']].dropna(
        subset=['formula_modelseed'])


def get_compared_formulae(formula_mismatch: pd.DataFrame) -> pd.DataFrame:
    """Compare formula by atom pattern

    Args:
        formula_mismatch (pd.DataFrame): 
            Table with column containing atom comparison. Output from :py:func:`~refinegems.utility.db_access.get_formula_mismatch`.

    Returns:
        pd.DataFrame: 
            table containing metabolites with formula mismatch
    """

    def formula_comparison(f1, f2):
        # from Jan Leusch
        formula_pattern = "[A-Z][a-z]?\\d*"
        atom_pattern = '[A-Z][a-z]?'
        atom_number_pattern = '\\d+'
        difference = {}
        f1_dict = {}
        f2_dict = {}
        f1 = re.findall(formula_pattern, f1)
        f2 = re.findall(formula_pattern, f2)
        for p in f1:
            key = re.findall(atom_pattern, p)[0]
            value = re.findall(atom_number_pattern, p)
            if not value:
                value = 1
            else:
                value = (int)(value[0])
            f1_dict[key] = value
        for q in f2:
            key = re.findall(atom_pattern, q)[0]
            value = re.findall(atom_number_pattern, q)
            if not value:
                value = 1
            else:
                value = (int)(value[0])
            f2_dict[key] = value
        difference = f1_dict
        if not f2_dict:
            return difference
        else:
            for q in f2_dict:
                if q in difference:
                    difference[q] -= f2_dict[q]
                else:
                    difference[q] = -f2_dict[q]
        for a in list(difference):
            if difference[a] == 0:
                difference.pop(a)
        return difference

    formula_mismatch['formula_comparison'] = formula_mismatch.apply(
        lambda row: formula_comparison(
            row['formula_model'], row['formula_modelseed']), axis=1)

    return formula_mismatch


def compare_to_modelseed(model: cobraModel) -> tuple[pd.DataFrame, pd.DataFrame]:
    """Executes all steps to compare model metabolites to ModelSEED metabolites

    Args:
        - model (cobraModel): 
            Model loaded with COBRApy

    Returns:
        tuple: 
            Tables with charge (1) & formula (2) mismatches

            (1) pd.DataFrame: Table with charge mismatches 
            (2) pd.DataFrame: Table with formula mismatches
    """
    ms_comp = get_modelseed_compounds()
    model_charges = get_model_charges(model)
    modelseed_charges = get_modelseed_charges(ms_comp)
    df_comp = compare_model_modelseed(model_charges, modelseed_charges)
    charge_mismatch = get_charge_mismatch(df_comp)
    formula_mismatch = get_formula_mismatch(df_comp)
    formula_comp = get_compared_formulae(formula_mismatch)
    return charge_mismatch, formula_comp



# NCBI
# ----
"""Connect to NBCI (using Biopython/Entrez) to extract information
like EC number of an NBCI protein accession number or information about locus tags.
"""

def search_ncbi_for_gpr(locus: str) -> str:
    """Fetches protein name from NCBI

    Args:
        - locus (str): 
            NCBI compatible locus_tag

    Returns:
        str: 
            Protein name|description
    """
    handle = Entrez.efetch(
        db="protein",
        id=locus,
        rettype="gbwithparts",
        retmode='text')
    records = SeqIO.parse(handle, "gb")

    for i, record in enumerate(records):
        if (locus[0] == 'W'):
            return record.description, locus
        else:
            for feature in record.features:
                if feature.type == "CDS":
                    return record.description, feature.qualifiers["locus_tag"][0]


# fetching the EC number (if possible) from NCBI 
# based on an NCBI protein ID (accession version number)
# @TODO logging
def get_ec_from_ncbi(mail:str,ncbiprot:str) -> str|None:
    """Based on a NCBI protein accession number, try and fetch the 
    EC number from NCBI.

    Args:
        - mail (str): 
            User's mail address for the NCBI ENtrez tool.
        - ncbiprot (str): 
            The NCBI protein accession number.

    Returns:
        (1) Case: fetching successful 
                str: 
                    The EC number associated with the protein ID based on NCBI.
        
        (2) Case: fetching unsuccessful
                None:
                    Nothing to return
     """
    try: 
        Entrez.email = mail
        handle = Entrez.efetch(db='protein', id=ncbiprot, rettype='gpc', retmode='xml')
        for feature_dict in xmltodict.parse(handle)['INSDSet']['INSDSeq']['INSDSeq_feature-table']['INSDFeature']:
            if isinstance(feature_dict['INSDFeature_quals']['INSDQualifier'], list):
                for qual in feature_dict['INSDFeature_quals']['INSDQualifier']:
                    if qual['INSDQualifier_name'] == 'EC_number':
                        return qual['INSDQualifier_value']
    except Exception as e:
        # @TODO : logging / warning etc. ?
        return None
  
  
# Uniprot
# -------
"""Currently mainly used for gapfilling, map your proteins against EC numbers 
using the SwissProt database"""

def map_dmnd_res_to_sp_ec_brenda(dmnd_results:pd.DataFrame, 
                                 swissprot_mapping_path:str) -> pd.DataFrame:
    """Map the results of a DIAMOND BLASTp run (filtered, see 
    :py:func:`~refinegems.curation.db_access.db.filter_DIAMOND_blastp_results`)

    Args:
        - dmnd_results (pd.DataFrame): 
            The results of the DIAMOND run.
        - swissprot_mapping_path (str): 
            The path to the SwissProt mapping file (IDs against BRENDA and EC,
            for information on how to get them, refer to :py:func:`~refinegems.utility.set_up.download_url`)

    Returns:
        pd.DataFrame: 
            The resulting mapping (no duplicates).
    """
    
    def _combine_EC_BRENDA(brenda:str,ec:str) -> list:
        """Helper function for combining the entries for BRENDA and EC number 
        for one ID (SwissProt mapping file) into a list of EC numbers.

        Args:
            - brenda (str): 
                The BRENDA column value.
            - ec (str): 
                The EC number column value.

        Returns:
            list: 
                The list with the EC numbers.
        """
    
        nums_list = []
        # get brenda ids
        if brenda and not pd.isna(brenda):
            for num in brenda.split(';'):
                nums_list.append(num.strip())

        # get ec numbers
        if ec and not pd.isna(ec):
            for num in ec.split(';'):
                nums_list.append(num.strip())
            
        # remove duplicates
        nums_list = list(set(nums_list))
        # filter non-complete EC numbers : X.X.X.X => 7 or more characters
        nums_list = [_ for _ in nums_list if len(_) >= 7]
            
        return nums_list
    
    # load the SwissProt mapping file
    swissprot_mapping = pd.read_csv(swissprot_mapping_path, sep='\t')
    swissprot_mapping.dropna(subset=['BRENDA','EC number'], how='all',inplace=True)
    
    # extract SwissProts IDs from subject_ID
    dmnd_results.columns = ['locus_tag','UniProt']
    dmnd_results['UniProt'] = dmnd_results['UniProt'].apply(lambda x: x.split('|')[1])
    
    # match
    dmnd_results = dmnd_results.merge(swissprot_mapping, left_on='UniProt', right_on='Entry', how='left')
    dmnd_results.drop('Entry', axis=1, inplace=True)
    dmnd_results['ec-code'] = dmnd_results.apply(lambda x: _combine_EC_BRENDA(x['BRENDA'],x['EC number']), axis=1)
    dmnd_results.drop(['BRENDA','EC number'], axis=1, inplace=True)
    dmnd_results = dmnd_results.explode('ec-code')
    dmnd_results.drop_duplicates(inplace=True)
    
    return dmnd_results

def get_ec_via_swissprot(fasta:str, db:str, missing_genes:pd.DataFrame, 
                         swissprot_mapping_file:str,
                         outdir:str=None,
                         sens:Literal['sensitive', 'more-sensitive', 'very-sensitive','ultra-sensitive']='more-sensitive',
                         cov:float=95.0,
                         t:int=2, pid:float=90.0) -> pd.DataFrame:
    """Based on a protein FASTA and a missing genes tables, mapped them to EC numbers 
    using a Swissprot DIAMOND database and a SwissProt mapping file (see :py:func:`~refinegems.utility.set_up.download_url`
    on how to download the needed files).

    Args:
        - fasta (str): 
            Path to the FASTA protein file.
        - db (str): 
            Path to the DIAMOND database (SwissProt).
        - missing_genes (pd.DataFrame): 
            The table of missing genes.
        - swissprot_mapping_file (str): 
            Path to the SwissProt mapping file.
        - outdir (str, optional): 
            Path to a directory to write the output to. 
            Defaults to None.
        - sens (Literal['sensitive', 'more-sensitive', 'very-sensitive','ultra-sensitive'], optional): 
            Sensitivity mode of DIAMOND blastp. Defaults to 'more-sensitive'.
        - cov (float, optional): 
            Coverage threshold for DIAMOND blastp. Defaults to 95.0.
        - t (int, optional): 
            Number of threads to use for DIAMOND blastp. Defaults to 2.
        - pid (float, optional): 
            Percentage identity value to use as a cutoff for the results
            of the DIAMOND blastp run. Defaults to 90.0.

    Returns:
        pd.DataFrame: 
            The missing genes table extended by the mapping to an EC number, 
            if successful.
    """
    
    # Step 1: Make a FASTA out of the missing genes
    miss_fasta = create_missing_genes_protein_fasta(fasta,outdir,missing_genes)
    # Step 2: Run DIAMOND
    #         blastp mode against SwissProt DB
    blast_path = run_DIAMOND_blastp(miss_fasta, db, 
                                    sensitivity=sens,
                                    coverage=cov,
                                    threads=t,
                                    outdir=outdir)
    # Step 3: filter DIAMOND hits
    dmnd_res = filter_DIAMOND_blastp_results(blast_path, pid)
    # Step 4: map to Swissprot mapping file
    mapped_res = map_dmnd_res_to_sp_ec_brenda(dmnd_res, swissprot_mapping_file)
    # Step 5: Aggregate UniProt IDs for unique combinations of 
    #         EC numbers and locus tags
    mapped_res = mapped_res.groupby(['locus_tag','ec-code']).agg({'UniProt': lambda x: x.tolist()}).reset_index()

    return mapped_res




#!/usr/bin/env python
"""This module provides functions for adding, handling and analysing 
the KEGG pathways (or more specific their annotations) contained in a model.
"""

__author__ = "Famke Baeuerle and Carolin Brune"

############################################################################
# requirements
############################################################################

import cobra
import re

from tqdm.auto import tqdm
from libsbml import SBMLReader, GroupsExtension
from libsbml import Model as libModel
from bioservices import KEGG
from ..utility.cvterms import add_cv_term_pathways, get_id_from_cv_term, add_cv_term_pathways_to_entity
from ..classes.reports import KEGGPathwayAnalysisReport

############################################################################
# functions
############################################################################

# adding KEGG reactions as Group Pathways
# ---------------------------------------
""" 
If your organism occurs in the KEGG database, 
extract the KEGG reaction ID from the annotations of your reactions and identify, 
in which KEGG pathways this reaction occurs. 
Add all KEGG pathways for a reaction then as annotations 
with the biological qualifier `OCCURS_IN` to the respective reaction.
"""

def load_model_enable_groups(modelpath: str) -> libModel:
    """Loads model as document using libSBML and enables groups extension

    Args:
        - modelpath (str): 
            Path to GEM

    Returns:
        libModel: 
            Model loaded with libSBML
    """
    reader = SBMLReader()
    read = reader.readSBMLFromFile(modelpath)  # read from file
    groupextension = GroupsExtension()
    groupURI = groupextension.getURI(3, 1, 1)
    read.enablePackage(groupURI, "groups", True)  # enable groups extension
    read.setPkgRequired('groups', False)  # make groups not required
    model = read.getModel()
    return model


def extract_kegg_reactions(model: libModel) -> tuple[dict, list]:
    """Extract KEGG Ids from reactions

    Args:
        - model (libModel): 
            Model loaded with libSBML. Output of :py:func:`~refinegems.curation.pathways.load_model_enable_groups`.

    Returns:
        tuple: 
            Dictionary 'reaction_id': 'KEGG_id' (1) & List of reactions without KEGG Id (2)
            
            (1) dict: Reaction Id as key and KEGG Id as value
            (2) list: Ids of reactions without KEGG annotation
    """
    list_reac = model.getListOfReactions()
    kegg_reactions = {}
    non_kegg_reac = []
    
    for reaction in list_reac:
        kegg_ids = get_id_from_cv_term(reaction, 'kegg.reaction')
        if len(kegg_ids) > 0:
            kegg_reactions[reaction.getId()] = kegg_ids[0]
        else:
            non_kegg_reac.append(reaction.getId())

    return kegg_reactions, non_kegg_reac


def extract_kegg_pathways(kegg_reactions: dict) -> dict:
    """Finds pathway for reactions in model with KEGG Ids, accesses KEGG API, uses tqdm to report progres to user

    Args:
        - kegg_reactions (dict): 
            Reaction Id as key and KEGG Id as value. Output[0] from :py:func:`~refinegems.curation.pathways.extract_kegg_reactions`.

    Returns:
        dict: 
            Reaction Id as key and KEGG Pathway Id as value
    """
    k = KEGG()
    kegg_pathways = {}

    print('Extracting pathway Id for each reaction:')
    for reaction in tqdm(kegg_reactions.keys()):
        kegg_reaction = k.get(kegg_reactions[reaction])
        dbentry = k.parse(kegg_reaction)
        # sometimes parse does not work -> try and except
        try:
            pathways = [x for x in dbentry['PATHWAY']]
        except BaseException:
            pathways = []
        kegg_pathways[reaction] = pathways

    return kegg_pathways


def add_kegg_pathways(model, kegg_pathways) -> libModel:
    """Add KEGG reactions as BQB_OCCURS_IN.

    Args:
        - model (libModel): 
            Model loaded with libSBML. Output of :py:func:`~refinegems.curation.pathways.load_model_enable_groups`.
        - kegg_pathways (dict): 
            Reaction Id as key and KEGG Pathway Id as value. Output of :py:func:`~refinegems.curation.pathways.extract_kegg_pathways`.

    Returns:
        libModel: 
        Modified model with KEGG pathways.
    """
    list_reac = model.getListOfReactions()

    for reaction in list_reac:
        if reaction.getId() in kegg_pathways.keys():
            for path in kegg_pathways[reaction.getId()]:
                add_cv_term_pathways_to_entity(path, 'KEGG', reaction)

    return model


def get_pathway_groups(kegg_pathways) -> dict:
    """Group reaction into pathways.

    Args:
        - kegg_pathways (dict): 
            Reaction Id as key and KEGG Pathway Id as value. Output of :py:func:`~refinegems.curation.pathways.extract_kegg_pathways`.

    Returns:
        dict: 
            KEGG Pathway Id as key and reactions Ids as values.
    """
    pathway_groups = {}
    for reaction in kegg_pathways.keys():
        for path in kegg_pathways[reaction]:
            if path not in pathway_groups.keys():
                pathway_groups[path] = [reaction]
            else:
                pathway_groups[path].append(reaction)
    return pathway_groups


def create_pathway_groups(model: libModel, pathway_groups) -> libModel:
    """Use group module to add reactions to KEGG pathway.

    Args:
        - model (libModel): 
            Model loaded with libSBML. Output of :py:func:`~refinegems.curation.pathways.load_model_enable_groups`.
        - pathway_groups (dict): 
            KEGG Pathway Id as key and reactions Ids as values. Output of :py:func:`~refinegmes.curation.pathways.get_pathway_groups`.

    Returns:
        libModel: 
            Modified model with groups for pathways.
    """
    k = KEGG()
    groups = model.getPlugin('groups')
    group_list = groups.getListOfGroups()
    keys = list(pathway_groups.keys())
    num_reactions = [len(sub) for sub in list(pathway_groups.values())]

    print('Adding pathways as groups to the model:')
    for i in tqdm(range(len(pathway_groups))):
        kegg_pathway = k.get(keys[i])
        dbentry = k.parse(kegg_pathway)
        if groups.getGroup('G_' + keys[i]) is not None:
            group = groups.getGroup('G_' + keys[i])
            group.setName(dbentry['NAME'][0])
            group.setMetaId("meta_" + 'G_' + keys[i])
            group.setKind('partonomy')
            group.setSBOTerm("SBO:0000633")  # NAME
            add_cv_term_pathways(keys[i], 'KEGG', group)
            for reac in pathway_groups[keys[i]]:
                if group.getMemberByIdRef(reac) is None:
                    member = group.createMember()
                    member.setIdRef(reac)
        else:
            group = group_list.createGroup() 
            group.setName(dbentry['NAME'][0])
            group.setId('G_' + keys[i])  # important for validity (memote/cobra)
            group.setMetaId("meta_" + 'G_' + keys[i])
            group.setKind('partonomy')
            group.setSBOTerm("SBO:0000633")  # NAME
            add_cv_term_pathways(keys[i], 'KEGG', group)
            for reac in pathway_groups[keys[i]]:
                if group.getMemberByIdRef(reac) is None:
                    member = group.createMember()
                    member.setIdRef(reac)

    return model


def kegg_pathways(modelpath: str) -> tuple[libModel, list[str]]:
    """Executes all steps to add KEGG pathways as groups

    Args:
        - modelpath (str): 
            Path to GEM.
        
    Returns:
        tuple: 
            libSBML model (1) & List of reactions without KEGG Id (2)

            (1) libModel: Modified model with Pathways as groups
            (2) list: Ids of reactions without KEGG annotation
    """
    model = load_model_enable_groups(modelpath)

    reactions, non_kegg_reactions = extract_kegg_reactions(model)
    pathways = extract_kegg_pathways(reactions)
    pathway_groups = get_pathway_groups(pathways)

    model_pathways = add_kegg_pathways(model, pathways)
    model_pathway_groups = create_pathway_groups(
        model_pathways, pathway_groups)
    
    return model_pathway_groups, non_kegg_reactions


# analyse the pathways in a model
# -------------------------------

def kegg_pathway_analysis(model:cobra.Model) -> KEGGPathwayAnalysisReport:
    """Analyse the pathways that are covered by the model.

    The analysis is based on the KEGG pathway classification and the available
    KEGG pathway identifiers present in the model.
    
    Note: one reaction can have multiple pathway identifiers associated with it. 
    This analysis focuses on the total number of IDs found within the model.

    Args:
        - model (cobra.Model): 
            A model loaded with COBRApy.

    Returns:
        KEGGPathwayAnalysisReport: 
            The KEGG pathway analysis report.
    """
    # create report
    report = KEGGPathwayAnalysisReport(total_reac=len(model.reactions))

    pathways = dict()
    counter = 0
    # extract KEGG pathway IDs from all reactions
    for r in model.reactions:
        if 'kegg.pathway' in r.annotation.keys():
            counter += 1
            anno = r.annotation['kegg.pathway']
            # case 1: only one annotation found
            if isinstance(anno,str):
                anno = re.sub(r'^[a-z]*','',anno)
                if anno in pathways:
                    pathways[anno] += 1
                else:
                    pathways[anno] = 1
            # case 2: multiple annotations for one reaction found
            else:
                for x in anno:
                    x = re.sub(r'^[a-z]*','',x)
                    if x in pathways:
                        pathways[x] += 1
                    else:
                        pathways[x] = 1
    
    # add counter to report
    report.kegg_count = counter

    # identify global and overview pathway identifier
    global_map = {}
    over_map = {}
    rest = {}
    for k,v in pathways.items():
        if k.startswith('011'):
            global_map[k] = v
        elif k.startswith('012'):
            over_map[k] = v
        else:
            rest[k] = v

    # add IDs in corresponding class to report
    report.kegg_global = global_map
    report.kegg_over = over_map
    report.kegg_paths = rest

    return report
